//
// This file is part of the imule Project.
//
// Copyright (c) 2003-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 1998 Vadim Zeitlin ( zeitlin@dptmaths.ens-cachan.fr )
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

#include <common/MuleDebug.h>			// Needed for MULE_VALIDATE_*
#include <common/StringFunctions.h>	// Needed for unicode2char
#include <common/Format.h>

#include "CFile.h"				// Interface declarations.
#include "FileFunctions.h"		// Needed for CheckFileExists
#include "Preferences.h"		// Needed for thePrefs
#include "Logger.h"				// Needed for AddDebugLogLineM

#include <unistd.h>				// Needed for close(2)
#include <cstdio>       		// SEEK_xxx constants
#include <fcntl.h>       		// O_RDONLY &c
#include <errno.h>				// errno

#include <wx/filefn.h>
#include <wx/log.h>				// Needed for wxSysErrorMsg
#include <wx/wfstream.h>
#include <wx/stream.h>

#ifdef HAVE_CONFIG_H
#include "config.h"             // Needed for HAVE_SYS_PARAM_H
#endif

// Mario Sergio Fujikawa Ferreira <lioux@FreeBSD.org>
// to detect if this is a *BSD system
#ifdef HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif


// This macro is used to check if a syscall failed, in that case
// log an appropriate message containing the errno string.
#define SYSCALL_CHECK(check, what) \
	do { \
		if (!(check)) { \
			AddDebugLogLineM(true, logCFile, \
				wxString() << wxT("Error when ") << what << wxT(" (") \
					<< m_filePath << wxT("): ") << wxSysErrorMsg()); \
		} \
	} while (false);


CSeekFailureException::CSeekFailureException(const wxString& desc)
	: CIOFailureException(wxT("SeekFailure"), desc)
{}


CFile::CFile()
	: wxFile()
{}


CFile::CFile(const wxString& fileName, OpenMode mode)
	: wxFile( fileName, (wxFile::OpenMode)mode)
{
	m_filePath = fileName ;
	m_mode = mode ;
	m_accessMode = s_defaultAccessMode ;
}


CFile::~CFile()
{ 
	if (IsOpened()) {
		Close(); 
	}
}


/*wxFile CFile::fd() const
{
	return m_fd;
}*/


bool CFile::IsOpened() const
{
	return wxFile::IsOpened();
}


const wxString& CFile::GetFilePath() const
{
	MULE_VALIDATE_STATE(IsOpened(), wxT("CFile: Cannot return path of closed file."));

	return m_filePath;
}


bool CFile::Create(const wxString& path, bool overwrite, int accessMode)
{
	if (!overwrite && CheckFileExists(path)) {
		return false;
	}

	return Open(path, write, accessMode);
}

int CFile::s_defaultAccessMode = wxS_DEFAULT;

bool CFile::Open(const wxString& fileName, OpenMode mode, int accessMode)
{
	MULE_VALIDATE_PARAMS(!fileName.IsEmpty(), wxT("CFile: Cannot open, empty path."));

	m_filePath = fileName ;
	m_mode = mode ;
	m_accessMode = accessMode ;
	
	return wxFile::Open( fileName, (wxFile::OpenMode)mode, accessMode );
}


bool CFile::Close() 
{
	MULE_VALIDATE_STATE(IsOpened(), wxT("CFile: Cannot close closed file."));

	return wxFile::Close();
}


bool CFile::Flush()
{
	MULE_VALIDATE_STATE(IsOpened(), wxT("CFile: Cannot flush closed file."));

	return wxFile::Flush();
}


size_t CFile::doRead(void* buffer, size_t count)
{
	MULE_VALIDATE_PARAMS(buffer, wxT("CFile: Invalid buffer in read operation."));
	MULE_VALIDATE_STATE(IsOpened(), wxT("CFile: Cannot read from closed file."));
	
	size_t totalRead = 0;
	while (totalRead < count) {
		size_t current = wxFile::Read( (char*) buffer + totalRead, count - totalRead) ;
		
		if (current == (size_t) wxInvalidOffset) {
			// Read error, nothing we can do other than abort.
			throw CIOFailureException(wxString(wxT("Error reading from file: ")) + wxSysErrorMsg());
		} else if ((totalRead + current < count) && Eof()) {
			// We may fail to read the specified count in a couple
			// of situations: EOF and interrupts. The check for EOF
			// is needed to avoid inf. loops.
			break;
		}
		
		totalRead += current;
	}
	
	return totalRead;
}

bool CFile::Eof() const
{
	return CFileDataIO::Eof();
}

size_t CFile::doWrite(const void* buffer, size_t nCount)
{
	MULE_VALIDATE_PARAMS(buffer, wxT("CFile: Invalid buffer in write operation."));
	MULE_VALIDATE_STATE(IsOpened(), wxT("CFile: Cannot write to closed file."));

	
	size_t result = wxFile::Write(buffer, nCount );
	
	if (result != nCount) {
		throw CIOFailureException(wxString(wxT("Error writing to file: ")) + wxSysErrorMsg());
	}
	return result;
}


wxULongLong CFile::doSeek(wxULongLong offset)
{
	MULE_VALIDATE_STATE(IsOpened(), wxT("Cannot seek on closed file."));
	
	wxULongLong result = wxFile::Seek( offset.GetValue(), wxFromStart );

	if (result == offset) {
		return result;
	} else if (result == (wxULongLong) wxInvalidOffset) {
		throw CSeekFailureException(wxString(wxT("Seeking failed: ")) + wxSysErrorMsg());
	} else {
		throw CSeekFailureException(wxT("Seeking returned incorrect position"));		
	}
}


wxULongLong CFile::GetPosition() const
{
	MULE_VALIDATE_STATE(IsOpened(), wxT("Cannot get position in closed file."));
	
	wxULongLong pos = wxFile::Tell();
	if (pos == (wxULongLong) wxInvalidOffset) {
		throw CSeekFailureException(wxString(wxT("Failed to retrieve position in file: ")) + wxSysErrorMsg());
	}
	
	return pos;
}


wxULongLong CFile::GetLength() const
{
	MULE_VALIDATE_STATE(IsOpened(), wxT("CFile: Cannot get length of closed file."));

	wxFileOffset l = wxFile::Length();
	
	if (l < (wxFileOffset) 0) {
		throw CIOFailureException(wxString(wxT("Failed to retrieve length of file: ")) + wxSysErrorMsg());
	}
	
	return wxULongLong(l);
}


bool CFile::SetLength( wxULongLong new_len )
{
	MULE_VALIDATE_STATE(IsOpened(), wxT("CFile: Cannot set length when no file is open."));

	wxULongLong written = 0 ;
	wxULongLong old_pos = GetPosition() ;
	wxULongLong old_len = GetLength()   ;

	Close() ;

	wxTempFileOutputStream f_out( m_filePath );

	if ( f_out.IsOk() )
	{
		wxFileInputStream f_in ( m_filePath );

		if (f_in.IsOk() )
		{
			wxBufferedOutputStream out( f_out );
			wxBufferedInputStream  in( f_in );
			
			while ( written < std::min(new_len, old_len ) )
			{
				out.PutC( (char) in.GetC() );
				written++;
			}
			while ( written < new_len )
			{
				out.PutC( 0 );
				written++;
			}
		}
		f_out.Commit();
	}

	Open( m_filePath, m_mode, m_accessMode );
	doSeek ( std::min(old_pos, written-wxULongLong(1)) );

	return ( written == new_len );
}
