//
// This file is part of the imule Project.
//
// Copyright (c) 2003-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2002 Merkur ( devs@emule-project.net / http://www.emule-project.net )
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

#ifndef CLIENTLIST_H
#define CLIENTLIST_H

#include "Types.h"		// Needed for uint16_t and uint32_t
#include "MD4Hash.h"		// Needed for CMD4Hash
#include "DeadSourceList.h"	// Needed for CDeadSourceList
#include "updownclient.h"   // Needed for states enum types
#include <deque>
#include <map>
#include <set>

#include "i2p/CI2PAddress.h"

class CUpDownClient;
class CClientTCPSocket;
class CDeletedClient;
class CMD4Hash;
namespace Kademlia {
	class CContact;
	class CUInt128;
}

/* NOTE: not for I2P
enum buddyState
{
	Disconnected,
	Connecting,
	Connected
};*/


#define BAN_CLEANUP_TIME	1200000 // 20 min


/**
 * This class takes care of managing existing clients.
 *
 * This class tracks a number of attributes related to existing and deleted
 * clients. Among other things, it keeps track of existing, banned, dead and
 * dying clients, as well as offers support for matching new client-instances
 * against already exist clients to avoid duplicates.
 */
class CClientList
{
public:
	/**
	 * Constructor.
	 */
	CClientList();

	/**
	 * Destructor.
	 */
	~CClientList();
	
	
	/**
	 * Adds a client to the global list of clients.
	 *
	 * @param toadd The new client.
	 */
	void	AddClient( CUpDownClient* toadd );

	/**
	 * Schedules a client for deletion.
	 *
	 * @param client The client to be deleted.
	 *
	 * Call this function whenever a client is to be deleted, rather than 
	 * directly deleting the client. If the client is on the global client-
	 * list, then it will be scheduled for deletion, otherwise it will be 
	 * deleted immediatly. Please check CUpDownClient::Safe_Delete for the
	 * proper way to do this.
	 */
	void	AddToDeleteQueue( CUpDownClient* client );


	/**
	 * Updates the recorded IP of the specified client.
	 *
	 * @param client The client to have its entry updated.
	 * @param newIP The new IP adress of the client.
	 *
	 * This function is to be called before the client actually changes its
	 * IP-address, and will update the old entry with the new value. There 
	 * will only be added an entry if the new IP isn't zero.
	 */
	//I2P:void	UpdateClientIP( CUpDownClient* client, uint32_t newIP );
	void	UpdateClientDest( CUpDownClient* client, const CI2PAddress & newDest );
	
	/**
	 * Updates the recorded hash of the specified client.
	 *
	 * @param client The client to have its entry updated.
	 * @param newHash The new user-hash.
	 *
	 * This function is to be called before the client actually changes its
	 * user-hash, and will update the old entry with the new value. There will 
	 * only be added an entry if the new hash is valid.
	 */
	void	UpdateClientHash( CUpDownClient* client, const CMD4Hash& newHash );


	/**
	 * Returns the number of listed clients.
	 */
	uint32_t	GetClientCount() const;

	
	/**
	 * Deletes all tracked clients.
	 */
	void	DeleteAll();


	/**
	 * Replaces a new client-instance with the an already existing client, if one such exist.
	 *
	 * @param client A pointer to the pointer of the new instance.
	 * @param sender The socket assosiated with the new instance.
	 *
	 * Call this function when a new client-instance has been created. This function will then 
	 * compare it against all existing clients and see if we already have an instance matching 
	 * the new one. If that is the case, it will delete the new instance and set the pointer to
	 * the existing one.
	 */
	bool	AttachToAlreadyKnown( CUpDownClient** client, CClientTCPSocket* sender );


	/**
	 * Finds a client with the specified ip and port.
	 *
	 * @param clientip The IP of the client to find.
	 * @param port The port used by the client.
	 */
	//I2P:CUpDownClient* FindClientByIP( uint32_t clientip, uint16_t port );
	CUpDownClient* FindClientByDest( const CI2PAddress & clientdest );


	// The list-type used to store clients IPs and other information
	//I2P:typedef std::map<uint32_t, uint32_t> ClientMap;
	
	//! The list-type used to store clients information
	typedef std::map<CI2PAddress, uint32_t> ClientMap;


	/**
	 * Adds a client to the list of tracked clients.
	 * 
	 * @param toadd The client to track.
	 *
	 * This function is used to keep track of clients after they 
	 * have been deleted and makes it possible to spot port or hash
	 * changes.
	 */
	void	AddTrackClient(CUpDownClient* toadd);

	/**
	 * Returns the number of tracked client.
	 *
	 * @param dwIP The IP-adress which of the clients.
	 * @return The number of clients tracked at the specifed IP.
	 */
	//uint16_t	GetClientsFromIP(uint32_t dwIP);
	uint16_t	GetClientsFromDest(const CI2PAddress & dwDest);

	/**
	 * Checks if a client has changed its user-hash.
	 * 
	 * @param dwIP The IP of the client.
	 * @param nPort The port of the client.
	 * @param pNewHash The userhash assosiated with the client.
	 * 
	 */
	// I2P	bool	ComparePriorUserhash( uint32_t dwIP, uint16_t nPort, void* pNewHash );
	bool	ComparePriorUserhash( const CI2PAddress & nDest, void* pNewHash );

	/**
	 * Bans an IP address for 2 hours.
	 *
	 * @param dwIP The IP from which all clients will be banned.
	 */
	//I2P: void	AddBannedClient(uint32_t dwIP);
	void	AddBannedClient(const CI2PAddress & nDest);
	
	/**
	 * Checks if a client has been banned.
	 *
	 * @param dwIP The IP to check.
	 * @return True if the IP is banned, false otherwise.
	 */
	//I2P:bool	IsBannedClient(uint32_t dwIP);
	bool	IsBannedClient(const CI2PAddress & nDest);
	
	/**
	 * Unbans an IP address, if it has been banned.
	 *
	 * @param dwIP The IP address to unban.
	 */
	//I2P:void	RemoveBannedClient(uint32_t dwIP);
	void	RemoveBannedClient(const CI2PAddress & nDest);


	/**
	 * Main loop.
	 *
	 * This function takes care of cleaning the various lists and deleting 
	 * pending clients on the deletion-queue.
	 */
	void	Process();


	/**
	 * Deletes clients previously queued for deletion
	 *
	 * This function takes care of deleting pending clients on the
	 * deletion-queue.
	 */
	void	ProcessDeleteQueue();


	/**
	 * This function removes all clients filtered by the current IPFilter.
	 *
	 * Call this function after changing the current IPFiler list, to ensure
	 * that no client-connections to illegal IPs exist. These would otherwise
	 * be allowed to exist, bypassing the IPFilter.
	 */
	void	FilterQueues();


	//! The type of the list used to store client-pointers for a couple of tasks.
	typedef std::deque<CUpDownClient*> SourceList;
	

	/**
	 * Returns a list of clients with the specified user-hash.
	 *
	 * @param hash The userhash to search for.
	 *
	 * This function will return a list of clients with the specified userhash,
	 * provided that the hash is a valid non-empty userhash. Empty hashes will
	 * simply result in nothing being found.
	 */
	SourceList	GetClientsByHash( const CMD4Hash& hash );
	
	/**
	 * Returns a list of clients with the specified IP.
	 *
	 * @param ip The IP-address to search for.
	 *
	 * This function will return a list of clients with the specified IP,
	 * provided that the IP is a non-zero value. A value of zero will not 
	 * result in any results.
	 */
	//SourceList	GetClientsByIP( unsigned long ip );
	SourceList	GetClientsByDest( const CI2PAddress & dest );


	
	//// The type of the lists used to store IDs.
	//typedef std::multimap<uint32_t, CUpDownClient*> IDMap;
	////! The pairs of the ID list.
	//typedef std::pair<uint32_t, CUpDownClient*> IDMapPair;
	
	//! The type of the lists used to store Addresses.
	typedef std::multimap<CI2PAddress, CUpDownClient*> IDMap;
	//! The pairs of the CI2PAddress list.
	typedef std::pair<CI2PAddress, CUpDownClient*> IDMapPair;

	/**
	 * Returns a list of all clients.
	 *
	 * @return The complete list of clients.
	 */
	const IDMap& GetClientList();


	/**
	 * Adds a source to the list of dead sources.
	 *
	 * @param client The source to be recorded as dead.
	 */
	void		AddDeadSource(const CUpDownClient* client);

	/**
	 * Checks if a source is recorded as being dead.
	 *
	 * @param client The client to evaluate.
	 * @return True if dead, false otherwise.
	 *
	 * Sources that are dead are not to be considered valid
	 * sources and should not be added to partfiles.
	 */
	bool		IsDeadSource(const CUpDownClient* client);
	
	/**
	 * Sends a message to a client, identified by a GUI_ID
	 *
	 * @return Success
	 */
	//I2P:bool	SendMessage(wxULongLong client_id, const wxString& message);
	bool	SendMessage(const CI2PAddress & client_dest, const wxString& message);
	 
	/**
	 * Stops a chat session with a client.
	 *
	 */
	 //void	SetChatState(wxULongLong client_id, uint8_t state);
	 void	SetChatState(const CI2PAddress & client_dest, EChatState state);

	/*
	 * Avoids unwanted clients to be forever in the client list
	 */
	void CleanUpClientList();

	//NOTE : not for I2P: uint8_t	GetBuddyStatus() const {return m_nBuddyStatus;}
	// This must be used on CreateKadSourceLink and if we ever add the columns
	// on shared files control.
	//NOTE : not for I2P: CUpDownClient* GetBuddy() const { return m_pBuddy; }
	void RequestTCP(Kademlia::CContact* contact);
	//NOTE : not for I2P: void RequestBuddy(Kademlia::CContact* contact);
	//NOTE : not for I2P: void IncomingBuddy(Kademlia::CContact* contact, Kademlia::CUInt128* buddyID );
	//NOTE : not for I2P: void RemoveFromKadList(CUpDownClient* torem);
	//NOTE : not for I2P: void AddToKadList(CUpDownClient* toadd);

private:
	/**
	 * Helperfunction which finds a client matching the specified client.
	 *
	 * @param client The client to search for.
	 * @return The matching client or NULL.
	 *
	 * This functions searches through the list of clients and finds the first match
	 * using the same checks as CUpDownClient::Compare, but without the overhead.
	 */
	CUpDownClient* FindMatchingClient( CUpDownClient* client );
	
	
	/**
	 * Helperfunction which removes the client from the dest-list.
	 */
	bool	RemoveFromList( CUpDownClient* client );
	/**
	 * Helperfunction which removes the client from the hash-list.
	 */
	void	RemoveHashFromList( CUpDownClient* client );


	//! The type of the list used to store user-hashes.
	typedef std::multimap<CMD4Hash, CUpDownClient*> HashMap;
	//! The pairs of the Hash-list.
	typedef std::pair<CMD4Hash, CUpDownClient*> HashMapPair;


	//! The map of clients with valid hashes
	HashMap	m_hashList;
	
	//! The map of clients
	IDMap	m_clientList;
	
	//! This is the lists of clients that should be deleted
	SourceList m_delete_queue;
#ifdef __WXDEBUG__
	bool m_delete_queue_closed;
#endif
	
	//! This is the map of banned clients.
	ClientMap m_bannedList;
	//! This variable is used to keep track of the last time the banned-list was pruned.
	uint32_t	m_dwLastBannCleanUp;

	//! This is the map of tracked clients.
	//std::map<uint32_t, CDeletedClient*> m_trackedClientsList;
	std::map<CI2PAddress, CDeletedClient*> m_trackedClientsList;
	
	//! This keeps track of the last time the tracked-list was pruned.
	uint32_t	m_dwLastTrackedCleanUp;

	//! This keeps track of the last time the client-list was pruned.
	uint32_t m_dwLastClientCleanUp;

	//! List of unusable sources.
	CDeadSourceList	m_deadSources;
	
	/* Kad Stuff */
	/*NOTE : not for I2P
	std::set<CUpDownClient*>	m_KadSources;
	CUpDownClient* m_pBuddy;
	uint8_t m_nBuddyStatus;
	*/
};

#endif
