//
// This file is part of the imule Project.
//
// Copyright (c) 2003-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2002 Merkur ( devs@emule-project.net / http://www.emule-project.net )
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

#include <wx/defs.h>		// Needed before any other wx/*.h

#include <wx/filefn.h>
#include <wx/textfile.h>	// Needed for wxTextFile
#include <wx/filename.h>
#include <wx/utils.h>

#include "DownloadQueue.h"	// Interface declarations
#include "Server.h"		// Needed for CServer
#include "Packet.h"		// Needed for CPacket
#include "MemFile.h"		// Needed for CMemFile
#include "ClientList.h"		// Needed for CClientList
#include "updownclient.h"	// Needed for CUpDownClient
#include "ServerList.h"		// Needed for CServerList
#include "ServerConnect.h"		// Needed for CServerConnect
#include "ED2KLink.h"		// Needed for CED2KFileLink
#include "SearchList.h"		// Needed for CSearchFile
#include "SharedFileList.h"	// Needed for CSharedFileList
#include "PartFile.h"		// Needed for CPartFile
#include "Preferences.h"	// Needed for thePrefs
#include "imule.h"			// Needed for theApp
#include "GetTickCount.h"	// Needed for GetTickCount
#include "AsyncDNS.h" // Needed for CAsyncDNS
#include "Statistics.h"		// Needed for theStats
#include "Logger.h"
#include <common/Format.h>		// Needed for CFormat
#include "IPFilter.h"
#include "FileFunctions.h"	// Needed for CDirIterator
#include "OtherFunctions.h"	// Needed for CMutexUnlocker
#include "FileLock.h"		// Needed for CFileLock
#include "GuiEvents.h"          // Needed for Nofify_... macros

#include "kademlia/kademlia/Kademlia.h"

#include <algorithm>
#include <numeric>
#include <vector>


// Max. file IDs per UDP packet
// ----------------------------
// 576 - 30 bytes of header (28 for UDP, 2 for "E3 9A" edonkey proto) = 546 bytes
// 546 / 16 = 34


#define MAX_FILES_PER_UDP_PACKET	31	// 2+16*31 = 498 ... is still less than 512 bytes!!
#define MAX_REQUESTS_PER_SERVER		35


CDownloadQueue::CDownloadQueue()
// Needs to be recursive that that is can own an observer assigned to itself
		: m_mutex ( wxMUTEX_RECURSIVE )
{
	m_datarate = 0;
	m_udpserver = 0;
	m_lastsorttime = 0;
	m_lastudpsearchtime = 0;
	m_lastudpstattime = 0;
	m_udcounter = 0;
	m_nLastED2KLinkCheck = 0;
	m_dwNextTCPSrcReq = 0;
	m_cRequestsSentToServer = 0;
	m_lastDiskCheck = 0;
	SetLastKademliaFileRequest();
}


CDownloadQueue::~CDownloadQueue()
{
	if ( !m_filelist.empty() )
	{
		for ( unsigned int i = 0; i < m_filelist.size(); i++ )
		{
			printf ( "\rSaving PartFile %u of %u", i + 1, ( unsigned int ) m_filelist.size() );
			fflush ( stdout );
			delete m_filelist[i];
		}

		printf ( "\nAll PartFiles Saved.\n" );
	}
}


void CDownloadQueue::LoadMetFiles ( const wxString& path )
{
	printf ( "Loading temp files from %s.\n",

	         ( const char * ) unicode2char ( path ) );

	CDirIterator TempDir ( path );

	std::vector<wxString> files;

	// Locate part-files to be loaded
	wxString fileName = TempDir.GetFirstFile ( CDirIterator::File, wxT ( "*.part.met" ) );

	while ( !fileName.IsEmpty() )
	{
		files.push_back ( fileName );

		fileName = TempDir.GetNextFile();
	}

	std::sort ( files.begin(), files.end() );

	// Load part-files

	for ( size_t i = 0; i < files.size(); i++ )
	{
		printf ( "\rLoading PartFile %u of %u", ( unsigned int ) ( i + 1 ), ( unsigned int ) files.size() );

		fileName = wxFileName ( files[i] ).GetFullName();

		CPartFile* toadd = new CPartFile();
		bool result = toadd->LoadPartFile ( path, fileName );

		if ( !result )
		{
			// Try from backup
			result = toadd->LoadPartFile ( path, fileName, true );
		}

		if ( result && !IsFileExisting ( toadd->GetFileHash() ) )
		{
			{

				wxMutexLocker lock ( m_mutex );

				m_filelist.push_back ( toadd );
			}

			NotifyObservers ( EventType ( EventType::INSERTED, toadd ) );

			if ( toadd->GetStatus ( true ) == PS_READY )
			{
				theApp.sharedfiles->SafeAddKFile ( toadd, true );
			}

			Notify_DownloadCtrlAddFile ( toadd );
		}
		else
		{
			delete toadd;

			wxString msg;

			if ( result )
			{
				msg << wxT ( "WARNING: Duplicate partfile with hash '" )
				<< toadd->GetFileHash().Encode() << wxT ( "' found, skipping: " )
				<< fileName;
			}
			else
			{
				msg << wxT ( "ERROR: Failed to load PartFile '" ) << fileName << wxT ( "'" );
			}

			AddDebugLogLineM ( true, logDownloadQueue, msg );

			// Newline so that the error stays visible.

			printf ( ": %s\n", ( const char* ) unicode2char ( msg ) );
		}
	}

	printf ( "\nAll PartFiles Loaded.\n" );

	if ( GetFileCount() == 0 )
	{
		AddLogLineM ( false, _ ( "No part files found" ) );
	}
	else
	{
		AddLogLineM ( false, wxString::Format ( _ ( "Found %u part files" ), GetFileCount() ) );

		DoSortByPriority();
		CheckDiskspace ( path );
	}
}


uint16_t CDownloadQueue::GetFileCount() const

{

	wxMutexLocker lock ( m_mutex );
	wxASSERT( m_filelist.size() <= (uint16_t) -1 );

	return (uint16_t) m_filelist.size();
}


CServer* CDownloadQueue::GetUDPServer() const

{

	wxMutexLocker lock ( m_mutex );

	return m_udpserver;
}


void CDownloadQueue::SetUDPServer ( CServer* server )

{

	wxMutexLocker lock ( m_mutex );

	m_udpserver = server;
}


void CDownloadQueue::SaveSourceSeeds()
{
	for ( uint16_t i = 0; i < GetFileCount(); i++ )
	{
		GetFileByIndex ( i )->SaveSourceSeeds();
	}
}


void CDownloadQueue::LoadSourceSeeds()
{
	for ( uint16_t i = 0; i < GetFileCount(); i++ )
	{
		GetFileByIndex ( i )->LoadSourceSeeds();
	}
}


void CDownloadQueue::AddSearchToDownload ( CSearchFile* toadd, uint8_t category )
{
	if ( IsFileExisting ( toadd->GetFileHash() ) )
	{
		return;
	}


	CPartFile* newfile = NULL;

	try
	{
		newfile = new CPartFile ( toadd );
	}
	catch ( const CInvalidPacket& e )
	{
		AddDebugLogLineM ( true, logDownloadQueue, wxT ( "Search-result contained invalid tags, could not add" ) );
	}

	if ( newfile && newfile->GetStatus() != PS_ERROR )
	{
		AddDownload ( newfile, thePrefs::AddNewFilesPaused(), category );
	}
	else
	{
		delete newfile;
	}
}


struct SFindBestPF
{
	void operator() ( CPartFile* file )
	{
		// Check if we should filter out other categories

		if ( ( m_category != -1 ) and ( file->GetCategory() != m_category ) )
		{
			return;
		}
		else if ( file->GetStatus() != PS_PAUSED )
		{
			return;
		}

		if ( !m_result or ( file->GetDownPriority() > m_result->GetDownPriority() ) )
		{
			m_result = file;
		}
	}

	//! The category to look for, or -1 if any category is good
	int m_category;
	//! If any acceptable files are found, this variable store their pointer
	CPartFile* m_result;
};


void CDownloadQueue::StartNextFile ( CPartFile* oldfile )
{
	if ( thePrefs::StartNextFile() )
	{
		SFindBestPF visitor = { -1, NULL };

		{

			wxMutexLocker lock ( m_mutex );

			if ( thePrefs::StartNextFileSame() )
			{
				// Get a download in the same category
				visitor.m_category = oldfile->GetCategory();

				visitor = std::for_each ( m_filelist.begin(), m_filelist.end(), visitor );
			}

			if ( visitor.m_result == NULL )
			{
				// Get a download, regardless of category
				visitor.m_category = -1;

				visitor = std::for_each ( m_filelist.begin(), m_filelist.end(), visitor );
			}
		}

		if ( visitor.m_result )
		{
			visitor.m_result->ResumeFile();
		}
	}
}


void CDownloadQueue::AddDownload ( CPartFile* file, bool paused, uint8_t category )
{
	wxCHECK_RET ( !IsFileExisting ( file->GetFileHash() ), wxT ( "Adding duplicate part-file" ) );

	if ( paused && GetFileCount() )
	{
		file->StopFile();
	}

	{

		wxMutexLocker lock ( m_mutex );

		m_filelist.push_back ( file );

		DoSortByPriority();
	}

	NotifyObservers ( EventType ( EventType::INSERTED, file ) );

	file->SetCategory ( category );
	Notify_DownloadCtrlAddFile ( file );
	AddLogLineM ( true, CFormat ( _ ( "Downloading %s" ) ) % file->GetFileName() );
}


bool CDownloadQueue::IsFileExisting ( const CMD4Hash& fileid ) const
{
	if ( CKnownFile* file = theApp.sharedfiles->GetFileByID ( fileid ) )
	{
		if ( file->IsPartFile() )
		{
			AddLogLineM ( true, CFormat ( _ ( "You are already trying to download the file '%s'" ) ) % file->GetFileName() );
		}
		else
		{
			// Check if the file exists, since otherwise the user is forced to
			// manually reload the shares to download a file again.
			wxString fullpath = JoinPaths ( file->GetFilePath(), file->GetFileName() );

			if ( !CheckFileExists ( fullpath ) )
			{
				// The file is no longer available, unshare it
				theApp.sharedfiles->RemoveFile ( file );

				return false;
			}

			AddLogLineM ( true, CFormat ( _ ( "You already have the file '%s'" ) ) % file->GetFileName() );
		}

		return true;
	}
	else if ( ( file = GetFileByID ( fileid ) ) )
	{
		AddLogLineM ( true, CFormat ( _ ( "You are already trying to download the file %s" ) ) % file->GetFileName() );
		return true;
	}

	return false;
}


void CDownloadQueue::Process()

{
	// send src requests to local server
	ProcessLocalRequests();

	{

		wxMutexLocker lock ( m_mutex );

		uint32_t downspeed = 0;

		if ( thePrefs::GetMaxDownload() != UNLIMITED && m_datarate > 1500 )
		{
			downspeed = ( ( ( uint32_t ) thePrefs::GetMaxDownload() ) * 1024 * 100 ) / ( m_datarate + 1 );

			if ( downspeed < 50 )
			{
				downspeed = 50;
			}
			else if ( downspeed > 200 )
			{
				downspeed = 200;
			}
		}

		m_datarate = 0;

		m_udcounter++;
		uint32_t cur_datarate = 0;
		uint8_t cur_udcounter = m_udcounter;

		for ( size_t i = 0; i < m_filelist.size(); i++ )
		{
			CPartFile* file = m_filelist[i];

			CMutexUnlocker unlocker ( m_mutex );

			if ( file->GetStatus() == PS_READY || file->GetStatus() == PS_EMPTY )
			{
				cur_datarate += file->Process ( downspeed, cur_udcounter );
			}
			else
			{
				//This will make sure we don't keep old sources to paused and stoped files..
				file->StopPausedFile();
			}
		}

		m_datarate += cur_datarate;


// 		if ( m_udcounter == 5 )
// 		{
// 			if ( theApp.serverconnect->IsUDPSocketAvailable() )
// 			{
// 				if ( ( ::GetTickCount() - m_lastudpstattime ) > UDPSERVERSTATTIME )
// 				{
// 					m_lastudpstattime = ::GetTickCount();
// 
// 					CMutexUnlocker unlocker ( m_mutex );
// 					theApp.serverlist->ServerStats();
// 				}
// 			}
// 		}
// 
// 		if ( m_udcounter == 10 )
// 		{
// 			m_udcounter = 0;
// 
// 			if ( theApp.serverconnect->IsUDPSocketAvailable() )
// 			{
// 				if ( ( ::GetTickCount() - m_lastudpsearchtime ) > UDPSERVERREASKTIME )
// 				{
// 					SendNextUDPPacket();
// 				}
// 			}
// 		}

		if ( ( ::GetTickCount() - m_lastsorttime ) > 10000 )
		{


			DoSortByPriority();
		}

		// Check if any paused files can be resumed

		CheckDiskspace ( thePrefs::GetTempDir() );
	}


	// Check for new links once per second.

	if ( ( ::GetTickCount() - m_nLastED2KLinkCheck ) >= 1000 )
	{
		AddLinksFromFile();
		m_nLastED2KLinkCheck = ::GetTickCount();
	}
}


CPartFile* CDownloadQueue::GetFileByID ( const CMD4Hash& filehash ) const
{

	wxMutexLocker lock ( m_mutex );

	for ( uint16_t i = 0; i < m_filelist.size(); ++i )
	{
		if ( filehash == m_filelist[i]->GetFileHash() )
		{
			return m_filelist[ i ];
		}
	}

	return NULL;
}


CPartFile* CDownloadQueue::GetFileByIndex ( unsigned int index )  const

{

	wxMutexLocker lock ( m_mutex );

	if ( index < m_filelist.size() )
	{
		return m_filelist[ index ];
	}

	wxASSERT ( false );

	return NULL;
}


bool CDownloadQueue::IsPartFile ( const CKnownFile* file ) const

{

	wxMutexLocker lock ( m_mutex );

	for ( uint16_t i = 0; i < m_filelist.size(); i++ )
	{
		if ( file == m_filelist[i] )
		{
			return true;
		}
	}

	return false;
}


bool CDownloadQueue::CheckAndAddSource ( CPartFile* sender, CUpDownClient* source )

{
	bool added = false;
	// if we block loopbacks at this point it should prevent us from connecting to ourself

	if ( source->HasValidHash() )
	{
		if ( source->GetUserHash() == thePrefs::GetUserHash() )
		{
			AddDebugLogLineM ( false, logDownloadQueue, wxT ( "Tried to add source with matching hash to your own." ) );
			source->Safe_Delete();
			return added;
		}
	}

	if ( sender->IsStopped() )
	{
		AddDebugLogLineM ( false, logDownloadQueue,
				   CFormat(wxT ( "Download of file %s is stopped. Source not added." ))
					   % sender->GetFullName() );
		source->Safe_Delete();
		return added;
	}

	// Filter sources which are known to be dead/useless
	if ( theApp.clientlist->IsDeadSource ( source ) || sender->IsDeadSource ( source ) )
	{
		AddDebugLogLineM ( false, logDownloadQueue,
				   CFormat( wxT ( "Dead source not added (udpdest=%s). Deleting." ) )
					 % source->GetUDPDest().humanReadable() );
		source->Safe_Delete();
		return added;
	}


	// Find all clients with the same hash
	if ( source->HasValidHash() )
	{
		CClientList::SourceList found = theApp.clientlist->GetClientsByHash ( source->GetUserHash() );

		CClientList::SourceList::iterator it = found.begin();

		for ( ; it != found.end(); it++ )
		{
			CKnownFile* file = ( *it )->GetRequestFile();
			// if the client is already requested for a file, delete the source, but add it as another request to the client
			// Only check files on the download-queue

			if ( file )
			{
				AddDebugLogLineM ( false, logDownloadQueue, wxT ( "Client has already been asked for a file" ) );
				// Is the found source queued for something else?

				if ( file != sender )
				{
					// Try to add a request for the other file
					AddDebugLogLineM ( false, logDownloadQueue, wxT ( "but another one" ) );

					if ( ( *it )->AddRequestForAnotherFile ( sender ) )
					{
						// Add it to downloadlistctrl
						AddDebugLogLineM ( false, logDownloadQueue, wxT ( "so let's put this as a pending request" ) );
						Notify_DownloadCtrlAddSource ( sender, *it, A4AF_SOURCE );
						added = true ;
					}
				}

				AddDebugLogLineM ( false, logDownloadQueue, wxT ( "... deleting the source" ) );

				source->Safe_Delete();
				return added;
			}
		}
	}



	// Our new source is real new but maybe it is already uploading to us?
	// If yes the known client will be attached to the var "source" and the old
	// source-client will be deleted. However, if the request file of the known
	// source is NULL, then we have to treat it almost like a new source and if
	// it isn't NULL and not "sender", then we shouldn't move it, but rather add
	// a request for the new file.
	ESourceFrom nSourceFrom = source->GetSourceFrom();

	if ( theApp.clientlist->AttachToAlreadyKnown ( &source, 0 ) )
	{
		// Already queued for another file?

		if ( source->GetRequestFile() )
		{
			// If we're already queued for the right file, then there's nothing to do
			AddDebugLogLineM ( false, logDownloadQueue, wxT ( "Client has already been asked for a file" ) );

			if ( sender != source->GetRequestFile() )
			{
				// Add the new file to the request list
				AddDebugLogLineM ( false, logDownloadQueue, wxT ( "... but another one. Add this." ) );
				source->AddRequestForAnotherFile ( sender );
				// added by mkvore :
				Notify_DownloadCtrlAddSource ( sender, source, UNAVAILABLE_SOURCE );
				 added = true ;
			}
		}
		else
		{
			// Source was known, but reqfile NULL.
			AddDebugLogLineM ( false, logDownloadQueue, wxT ( "Client already known. Set this file as a request." ) );
			source->SetRequestFile ( sender );

			if ( source->GetSourceFrom() != nSourceFrom )
			{
				if ( source->GetSourceFrom() != SF_NONE )
				{
					theStats::RemoveSourceOrigin ( source->GetSourceFrom() );
					theStats::RemoveFoundSource();
				}

				source->SetSourceFrom ( nSourceFrom );
			}

			sender->AddSource ( source );

			if ( source->GetFileRating() || !source->GetFileComment().IsEmpty() )
			{
				sender->UpdateFileRatingCommentAvail();
			}

			Notify_DownloadCtrlAddSource ( sender, source, UNAVAILABLE_SOURCE );
			added = true ;
		}
	}
	else
	{
		// Unknown client, add it to the clients list
		AddDebugLogLineM ( false, logDownloadQueue, wxT ( "Unknown client, add it to the client list." ) );
		source->SetRequestFile ( sender );

		theApp.clientlist->AddClient ( source );

		sender->AddSource ( source );

		if ( source->GetFileRating() || !source->GetFileComment().IsEmpty() )
		{
			sender->UpdateFileRatingCommentAvail();
		}

		Notify_DownloadCtrlAddSource ( sender, source, UNAVAILABLE_SOURCE );
		added = true ;
	}
	return added ;
}


void CDownloadQueue::CheckAndAddKnownSource ( CPartFile* sender, CUpDownClient* source )
{
	// Kad reviewed

	if ( sender->IsStopped() )
	{
		return;
	}

	// Filter sources which are known to be dead/useless
	if ( sender->IsDeadSource ( source ) )
	{
		return;
	}

	// "Filter LAN IPs" -- this may be needed here in case we are connected to the internet and are also connected
	// to a LAN and some client from within the LAN connected to us. Though this situation may be supported in future
	// by adding that client to the source list and filtering that client's LAN IP when sending sources to
	// a client within the internet.
	//
	// "IPfilter" is not needed here, because that "known" client was already IPfiltered when receiving OP_HELLO.
	//if ( !source->HasLowID() )
	{
		CI2PAddress nClientDest = source->GetTCPDest();

		if ( !IsGoodDest ( nClientDest ) )
		{ // check for 0-IP
			AddDebugLogLineM ( false, logDownloadQueue/*logIPFilter*/, CFormat ( wxT ( "Ignored already known source with dest=%x" ) ) % nClientDest.hashCode() );//Uint32_ttoStringIP(nClientIP));
			return;
		}
	}

	CPartFile* file = source->GetRequestFile();

	// Check if the file is already queued for something else

	if ( file )
	{
		if ( file != sender )
		{
			if ( source->AddRequestForAnotherFile ( sender ) )
			{
				Notify_DownloadCtrlAddSource ( sender, source, A4AF_SOURCE );
			}
		}
	}
	else
	{
		source->SetRequestFile ( sender );

		if ( source->GetFileRating() || !source->GetFileComment().IsEmpty() )
		{
			sender->UpdateFileRatingCommentAvail();
		}

		source->SetSourceFrom ( SF_PASSIVE );

		sender->AddSource ( source );
		Notify_DownloadCtrlAddSource ( sender, source, UNAVAILABLE_SOURCE );
	}
}


/**
 * Remove a client from the download queue
 * @param toremove the client to remove
 * @param WXUNUSED(updatewindow ) 
 * @param bDoStatsUpdate true if the stats have to be updated
 * @return 
 */
bool CDownloadQueue::RemoveSource ( CUpDownClient* toremove, bool	WXUNUSED ( updatewindow ), bool bDoStatsUpdate )
{
	bool removed = false;
	toremove->DeleteAllFileRequests();

	for ( uint16_t i = 0; i < GetFileCount(); i++ )
	{
		CPartFile* cur_file = GetFileByIndex ( i );

		// Remove from source-list

		if ( cur_file->DelSource ( toremove ) )
		{
			cur_file->RemoveDownloadingSource ( toremove );
			removed = true;

			if ( bDoStatsUpdate )
			{
				cur_file->UpdatePartsInfo();
			}
		}

		// Remove from A4AF-list
		cur_file->A4AFsrclist.erase ( toremove );
	}


	if ( !toremove->GetFileComment().IsEmpty() || toremove->GetFileRating() > 0 )
	{
		toremove->GetRequestFile()->UpdateFileRatingCommentAvail();
	}

	toremove->SetDownloadState ( DS_NONE );

	toremove->SetRequestFile ( NULL );

	// Remove from downloadlist widget
	Notify_DownloadCtrlRemoveSource ( toremove, 0 );
	toremove->ResetFileStatusInfo();

	return removed;
}


void CDownloadQueue::RemoveFile ( CPartFile* file )
{
	RemoveLocalServerRequest ( file );

	NotifyObservers ( EventType ( EventType::REMOVED, file ) );

	wxMutexLocker lock ( m_mutex );

	EraseValue ( m_filelist, file );
}


CUpDownClient* CDownloadQueue::GetDownloadClientByDest_UDP ( const CI2PAddress & nUDPDest ) const
{

	wxMutexLocker lock ( m_mutex );

	for ( unsigned int i = 0; i < m_filelist.size(); i++ )
	{
		const CPartFile::SourceSet& set = m_filelist[i]->GetSourceList();

		for ( CPartFile::SourceSet::iterator it = set.begin(); it != set.end(); it++ )
		{
			if ( ( *it )->GetUDPDest() == nUDPDest )
			{
				return *it;
			}
		}
	}

	return NULL;
}


/**
 * Checks if the specified server is the one we are connected to.
 */
bool IsConnectedServer ( const CServer* server )
{
	if ( server && theApp.serverconnect->GetCurrentServer() )
	{
		CI2PAddress srvDest = theApp.serverconnect->GetCurrentServer()->GetDest();
		//uint16_t srvPort = theApp.serverconnect->GetCurrentServer()->GetPort();

		return server->GetDest() == srvDest;//Addr && server->GetPort() == srvPort;
	}

	return false;
}


bool CDownloadQueue::SendNextUDPPacket()

{
	if ( m_filelist.empty() || !theApp.serverconnect->IsUDPSocketAvailable() || !theApp.IsConnectedED2K() )
	{
		return false;
	}

	// Start monitoring the server and the files list
	if ( !m_queueServers.IsActive() )
	{
		AddObserver ( &m_queueFiles );

		theApp.serverlist->AddObserver ( &m_queueServers );
	}


	bool packetSent = false;

	while ( !packetSent )
	{
		// Get max files ids per packet for current server
		int filesAllowed = GetMaxFilesPerUDPServerPacket();

		if ( filesAllowed < 1 || !m_udpserver || IsConnectedServer ( m_udpserver ) )
		{
			// Select the next server to ask, must not be the connected server

			do
			{
				m_udpserver = m_queueServers.GetNext();
			}
			while ( IsConnectedServer ( m_udpserver ) );

			m_cRequestsSentToServer = 0;

			filesAllowed = GetMaxFilesPerUDPServerPacket();
		}


		// Check if we have asked all servers, in which case we are done
		if ( m_udpserver == NULL )
		{
			DoStopUDPRequests();

			return false;
		}

		// Memoryfile containing the hash of every file to request
		// 20bytes allocation because 16b + 4b is the worse case scenario.
		CMemFile hashlist ( 20 );

		CPartFile* file = m_queueFiles.GetNext();

		while ( file && filesAllowed )
		{
			uint8_t status = file->GetStatus();

			if ( ( status == PS_READY || status == PS_EMPTY ) && file->GetSourceCount() < thePrefs::GetMaxSourcePerFileUDP() )
			{
				hashlist.WriteHash ( file->GetFileHash() );

				if ( m_udpserver->GetUDPFlags() & SRV_UDPFLG_EXT_GETSOURCES2 )
				{
					hashlist.WriteUInt64 ( file->GetFileSize().GetValue() );
				}

				--filesAllowed;
			}

			// Avoid skipping a file if we can't send any more currently
			if ( filesAllowed )
			{
				file = m_queueFiles.GetNext();
			}
		}

		// See if we have anything to send
		if ( hashlist.GetLength() != (wxULongLong) 0 )
		{
			packetSent = SendGlobGetSourcesUDPPacket ( hashlist );
		}


		// Check if we've covered every file
		if ( file == NULL )
		{
			// Reset the list of asked files so that the loop will start over
			m_queueFiles.Reset();


			// Unset the server so that the next server will be used
			m_udpserver = NULL;
		}
	}

	return true;
}


void CDownloadQueue::StopUDPRequests()
{

	wxMutexLocker lock ( m_mutex );

	DoStopUDPRequests();
}


void CDownloadQueue::DoStopUDPRequests()
{
	// No need to observe when we wont be using the results
	theApp.serverlist->RemoveObserver ( &m_queueServers );
	RemoveObserver ( &m_queueFiles );

	m_udpserver = 0;
	m_lastudpsearchtime = ::GetTickCount();
}


// Comparison function needed by sort. Returns true if file1 preceeds file2

bool ComparePartFiles ( const CPartFile* file1, const CPartFile* file2 )
{
	if ( file1->GetDownPriority() != file2->GetDownPriority() )
	{
		// To place high-priority files before low priority files we have to
		// invert this test, since PR_LOW is lower than PR_HIGH, and since
		// placing a PR_LOW file before a PR_HIGH file would mean that
		// the PR_LOW file gets sources before the PR_HIGH file ...
		return ( file1->GetDownPriority() > file2->GetDownPriority() );
	}
	else
	{
		int sourcesA = file1->GetSourceCount();
		int sourcesB = file2->GetSourceCount();

		int notSourcesA = file1->GetNotCurrentSourcesCount();
		int notSourcesB = file2->GetNotCurrentSourcesCount();

		int cmp = CmpAny ( sourcesA - notSourcesA, sourcesB - notSourcesB );

		if ( cmp == 0 )
		{
			cmp = CmpAny ( notSourcesA, notSourcesB );
		}

		return cmp < 0;
	}
}


void CDownloadQueue::DoSortByPriority()

{
	m_lastsorttime = ::GetTickCount();
	sort ( m_filelist.begin(), m_filelist.end(), ComparePartFiles );
}


void CDownloadQueue::ResetLocalServerRequests()
{

	wxMutexLocker lock ( m_mutex );

	m_dwNextTCPSrcReq = 0;

	m_localServerReqQueue.clear();

	for ( uint16_t i = 0; i < m_filelist.size(); i++ )
	{
		m_filelist[i]->m_bLocalSrcReqQueued = false;
	}
}


void CDownloadQueue::RemoveLocalServerRequest ( CPartFile* file )
{

	wxMutexLocker lock ( m_mutex );

	EraseValue ( m_localServerReqQueue, file );

	file->m_bLocalSrcReqQueued = false;
}


void CDownloadQueue::ProcessLocalRequests()
{

	wxMutexLocker lock ( m_mutex );

	if ( ( !m_localServerReqQueue.empty() ) && ( m_dwNextTCPSrcReq < ::GetTickCount() ) )
	{
		CMemFile dataTcpFrame ( 22 );

		const int iMaxFilesPerTcpFrame = 15;

		int iFiles = 0;

		while ( !m_localServerReqQueue.empty() && iFiles < iMaxFilesPerTcpFrame )
		{
			// find the file with the longest waitingtime
			uint32_t dwBestWaitTime = 0xFFFFFFFF;

			std::list<CPartFile*>::iterator posNextRequest = m_localServerReqQueue.end();
			std::list<CPartFile*>::iterator it = m_localServerReqQueue.begin();

			while ( it != m_localServerReqQueue.end() )
			{
				CPartFile* cur_file = ( *it );

				if ( cur_file->GetStatus() == PS_READY || cur_file->GetStatus() == PS_EMPTY )
				{
					uint8_t nPriority = cur_file->GetDownPriority();

					if ( nPriority > PR_HIGH )
					{
						wxASSERT ( 0 );
						nPriority = PR_HIGH;
					}

					if ( cur_file->lastsearchtime + ( PR_HIGH - nPriority ) < dwBestWaitTime )
					{
						dwBestWaitTime = cur_file->lastsearchtime + ( PR_HIGH - nPriority );
						posNextRequest = it;
					}

					it++;
				}
				else
				{
					it = m_localServerReqQueue.erase ( it );
					cur_file->m_bLocalSrcReqQueued = false;
					AddDebugLogLineM ( false, logDownloadQueue, wxString ( wxT ( "Local server source request for file \"" ) ) + cur_file->GetFileName() + wxString ( wxT ( "\" not sent because of status '" ) ) +  cur_file->getPartfileStatus() + wxT ( "'" ) );
				}
			}

			if ( posNextRequest != m_localServerReqQueue.end() )
			{
				CPartFile* cur_file = ( *posNextRequest );
				cur_file->m_bLocalSrcReqQueued = false;
				cur_file->lastsearchtime = ::GetTickCount();
				m_localServerReqQueue.erase ( posNextRequest );
				iFiles++;

				// create request packet
				CMemFile data ( 20 );
				data.WriteHash ( cur_file->GetFileHash() );
				// Kry - lugdunum extended protocol on 17.3 to handle filesize properly.
				// There is no need to check anything, old server ignore the extra 4 bytes.
				data.WriteUInt64 ( cur_file->GetFileSize().GetValue() );
				CPacket packet ( &data );
				packet.SetOpCode ( OP_GETSOURCES );
				dataTcpFrame.Write ( packet.GetPacket(), packet.GetRealPacketSize() );
			}
		}

		int iSize = (int) dataTcpFrame.GetLength().GetValue();

		if ( iSize > 0 )
		{
			// create one 'packet' which contains all buffered OP_GETSOURCES eD2K packets to be sent with one TCP frame
			// server credits: 16*iMaxFilesPerTcpFrame+1 = 241
			CPacket* packet = new CPacket ( new char[iSize], iSize, true, false );
			dataTcpFrame.Seek ( 0, wxFromStart );
			dataTcpFrame.Read ( packet->GetPacket(), iSize );
			uint32_t size = packet->GetPacketSize();
			theApp.serverconnect->SendPacket ( packet, true );	// Deletes `packet'.
			theStats::AddUpOverheadServer ( size );
		}

		// next TCP frame with up to 15 source requests is allowed to be sent in..
		m_dwNextTCPSrcReq = ::GetTickCount() + SEC2MS ( iMaxFilesPerTcpFrame * ( 16 + 4 ) );
	}

}


void CDownloadQueue::SendLocalSrcRequest ( CPartFile* sender )
{

	wxMutexLocker lock ( m_mutex );

	m_localServerReqQueue.push_back ( sender );
}


void CDownloadQueue::AddLinksFromFile()
{
	const wxString fullPath = theApp.ConfigDir + wxT ( "ED2KLinks" );

	if ( not wxFile::Exists ( fullPath ) )
	{
		return;
	}

	// Attempt to lock the ED2KLinks file.
	CFileLock lock ( ( const char* ) unicode2char ( fullPath ) );

	wxTextFile file ( fullPath );

	if ( file.Open() )
	{
		for ( unsigned int i = 0; i < file.GetLineCount(); i++ )
		{
			wxString line = file.GetLine ( i ).Strip ( wxString::both );

			if ( !line.IsEmpty() )
			{
				// Special case! used by a secondary running mule to raise this one.

				if ( line == wxT ( "RAISE_DIALOG" ) )
				{
					Notify_ShowGUI();
					continue;
				}

				AddED2KLink ( line );
			}
		}

		file.Close();
	}
	else
	{
		printf ( "Failed to open ED2KLinks file.\n" );
	}

	// Delete the file.
	wxRemoveFile ( theApp.ConfigDir +  wxT ( "ED2KLinks" ) );
}


void CDownloadQueue::ResetCatParts ( uint8_t cat )
{
	for ( uint16_t i = 0; i < GetFileCount(); i++ )
	{
		CPartFile* file = GetFileByIndex ( i );

		if ( file->GetCategory() == cat )
		{
			// Reset the category
			file->SetCategory ( 0 );
		}
		else if ( file->GetCategory() > cat )
		{
			// Set to the new position of the original category
			file->SetCategory ( (uint8_t) (file->GetCategory() - 1) );
		}
	}
}


void CDownloadQueue::SetCatPrio ( uint8_t cat, uint8_t newprio )
{
	for ( uint16_t i = 0; i < GetFileCount(); i++ )
	{
		CPartFile* file = GetFileByIndex ( i );

		if ( !cat || file->GetCategory() == cat )
		{
			if ( newprio == PR_AUTO )
			{
				file->SetAutoDownPriority ( true );
			}
			else
			{
				file->SetAutoDownPriority ( false );
				file->SetDownPriority ( newprio );
			}
		}
	}
}


void CDownloadQueue::SetCatStatus ( uint8_t cat, int newstatus )
{
	std::list<CPartFile*> files;

	{

		wxMutexLocker lock ( m_mutex );

		for ( uint16_t i = 0; i < m_filelist.size(); i++ )
		{
			if ( m_filelist[i]->CheckShowItemInGivenCat ( cat ) )
			{
				files.push_back ( m_filelist[i] );
			}
		}
	}

	std::list<CPartFile*>::iterator it = files.begin();

	for ( ; it != files.end(); it++ )
	{
		switch ( newstatus )
		{
			case MP_CANCEL:	( *it )->Delete(); 	break;

			case MP_PAUSE:	( *it )->PauseFile();	break;

			case MP_STOP:	( *it )->StopFile();	break;

			case MP_RESUME:	( *it )->ResumeFile();	break;
		}
	}
}


uint16_t CDownloadQueue::GetDownloadingFileCount() const
{

	wxMutexLocker lock ( m_mutex );

	uint16_t count = 0;

	for ( uint16_t i = 0; i < m_filelist.size(); i++ )
	{
		uint8_t status = m_filelist[i]->GetStatus();

		if ( status == PS_READY || status == PS_EMPTY )
		{
			count++;
		}
	}

	return count;
}


uint16_t CDownloadQueue::GetPausedFileCount() const

{

	wxMutexLocker lock ( m_mutex );

	uint16_t count = 0;

	for ( uint16_t i = 0; i < m_filelist.size(); i++ )
	{
		if ( m_filelist[i]->GetStatus() == PS_PAUSED )
		{
			count++;
		}
	}

	return count;
}


void CDownloadQueue::CheckDiskspace ( const wxString& path )

{
	if ( ::GetTickCount() - m_lastDiskCheck < DISKSPACERECHECKTIME )
	{
		return;
	}

	m_lastDiskCheck = ::GetTickCount();

	int min = 0;
	// Check if the user has set an explicit limit

	if ( thePrefs::IsCheckDiskspaceEnabled() )
	{
		min = thePrefs::GetMinFreeDiskSpace();
	}

	// The very least acceptable diskspace is a single PART
	if ( min < PARTSIZE )
	{
		min = PARTSIZE;
	}

	// Get the current free disc-space
	wxLongLong free = (wxLongLong) 0;

	if ( wxGetDiskSpace ( path, NULL, &free ) )
	{
		return;
	}


	for ( size_t i = 0; i < m_filelist.size(); i++ )
	{
		CPartFile* file = m_filelist[i];

		switch ( file->GetStatus() )
		{
			case PS_ERROR:
			case PS_COMPLETING:
			case PS_COMPLETE:
				continue;
		}

		if ( free >= wxLongLong(min) && file->GetInsufficient() )
		{
			// We'll try to resume files if there is enough free space

			if ( (wxULongLong) (wxULongLong_t) free.GetValue() > file->GetNeededSpace() + (wxULongLong) min )
			{
				file->ResumeFile();
			}
		}
		else if ( free < min && !file->IsPaused() )
		{
			// No space left, stop the files.
			file->PauseFile ( true );
		}
	}
}


int CDownloadQueue::GetMaxFilesPerUDPServerPacket() const

{
	if ( m_udpserver )
	{
		if ( m_udpserver->GetUDPFlags() & SRV_UDPFLG_EXT_GETSOURCES )
		{
			// get max. file ids per packet

			if ( m_cRequestsSentToServer < MAX_REQUESTS_PER_SERVER )
			{
				return std::min (
				           MAX_FILES_PER_UDP_PACKET,
				           MAX_REQUESTS_PER_SERVER - m_cRequestsSentToServer
				       );
			}
		}
		else if ( m_cRequestsSentToServer < MAX_REQUESTS_PER_SERVER )
		{
			return 1;
		}
	}

	return 0;
}


bool CDownloadQueue::SendGlobGetSourcesUDPPacket ( CMemFile& data )

{
	if ( !m_udpserver )
	{
		return false;
	}

	int item_size;

	if ( m_udpserver->GetUDPFlags() & SRV_UDPFLG_EXT_GETSOURCES2 )
	{
		item_size = ( 16 + 4 ); // (hash + size)
	}
	else
	{
		item_size = 16;
	}

	size_t iFileIDs = (size_t) data.GetLength().GetValue() / item_size;

	CPacket packet ( &data );

	if ( item_size == 16 )
	{
		packet.SetOpCode ( OP_GLOBGETSOURCES );
	}
	else
	{
		packet.SetOpCode ( OP_GLOBGETSOURCES2 );
	}

	theStats::AddUpOverheadServer ( packet.GetPacketSize() );

	theApp.serverconnect->SendUDPPacket ( &packet, m_udpserver, false );

	wxASSERT( m_cRequestsSentToServer + iFileIDs <= (size_t) (uint8_t) (-1) );
	m_cRequestsSentToServer = (uint8_t) (m_cRequestsSentToServer + iFileIDs);
	return true;
}


void CDownloadQueue::AddToResolve ( const CMD4Hash& fileid, const /*wxString*/ CI2PAddress & pszHostname ) //const wxString& pszHostname)//, )//uint16_t port)
{
	// double checking

	if ( !GetFileByID ( fileid ) )
	{
		return;
	}

	wxMutexLocker lock ( m_mutex );

	Hostname_Entry entry = { fileid, pszHostname };//, port };

	m_toresolve.push_front ( entry );

	// Check if there are other DNS lookups on queue
	if ( m_toresolve.size() == 1 )
	{
		// Check if it is a simple dot address
		/*TODO: check if this is a destination
		uint32_t ip = StringIPtoUint32(pszHostname);

		if (ip) {
			OnHostnameResolved(ip);
		*/
		//CI2PAddress dest = CI2PAddress::fromString(pszHostname);
		CI2PAddress dest = pszHostname;

		if ( dest.isValid() )
		{
			OnHostnameResolved ( dest );
		}
		else
		{
			CAsyncDNS* dns = new CAsyncDNS ( dest.getAlias() /*pszHostname*/, DNS_SOURCE, &theApp );

			if ( ( dns->Create() != wxTHREAD_NO_ERROR ) || ( dns->Run() != wxTHREAD_NO_ERROR ) )
			{
				dns->Delete();
				m_toresolve.pop_front();
			}
		}
	}
}


void CDownloadQueue::OnHostnameResolved ( const CI2PAddress & dest )

{

	wxMutexLocker lock ( m_mutex );

	wxASSERT ( m_toresolve.size() );

	AddDebugLogLineM ( false, logDownloadQueue, CFormat ( wxT ( "OnHostnameResolved %x" ) ) % dest.hashCode() );

	Hostname_Entry resolved = m_toresolve.front();

	m_toresolve.pop_front();

	if ( dest.isValid() )
	{
		CPartFile* file = GetFileByID ( resolved.fileid );

		if ( file )
		{
			CMemFile sources ( 1 + dest.size() );//4+2);
			sources.WriteUInt8 ( 1 ); // No. Sources
			sources.WriteAddress ( dest );

			sources.Seek ( 0, wxFromStart );

			file->AddSources ( sources, CI2PAddress::null/*0, 0*/, SF_LINK );
		}
	}

	while ( m_toresolve.size() )
	{
		Hostname_Entry entry = m_toresolve.front();

		// Check if it is a simple dot address
		//uint32_t tmpIP = StringIPtoUint32(entry.strHostname);
		CI2PAddress tmpDest = entry.strHostname;

		if ( tmpDest.isValid() )
		{
			OnHostnameResolved ( tmpDest );
		}
		else
		{
			CAsyncDNS* dns = new CAsyncDNS ( entry.strHostname.getAlias(), DNS_SOURCE, &theApp );

			if ( ( dns->Create() != wxTHREAD_NO_ERROR ) || ( dns->Run() != wxTHREAD_NO_ERROR ) )
			{
				dns->Delete();
				m_toresolve.pop_front();
			}
			else
			{
				break;
			}
		}
	}
}


bool CDownloadQueue::AddED2KLink ( const wxString& link, uint8_t category )

{
	wxASSERT ( !link.IsEmpty() );
	wxString URI = link;

	// Need the links to end with /, otherwise CreateLinkFromUrl crashes us.

	if ( URI.Last() != wxT ( '/' ) )
	{
		URI += wxT ( "/" );
	}

	try
	{
		CED2KLink* uri = CED2KLink::CreateLinkFromUrl ( URI );
		bool result = AddED2KLink ( uri, category );
		delete uri;

		return result;
	}
	catch ( const wxString& err )
	{
		AddLogLineM ( true, CFormat ( _ ( "Invalid ed2k link! Error: %s" ) ) % err );
	}

	return false;
}


bool CDownloadQueue::AddED2KLink ( const CED2KLink* link, uint8_t category )

{
	switch ( link->GetKind() )
	{
		case CED2KLink::kFile:
			return AddED2KLink ( dynamic_cast<const CED2KFileLink*> ( link ), category );

		case CED2KLink::kServer:
			return AddED2KLink ( dynamic_cast<const CED2KServerLink*> ( link ) );

		case CED2KLink::kServerList:
			return AddED2KLink ( dynamic_cast<const CED2KServerListLink*> ( link ) );

		default:
			return false;
	}
}



bool CDownloadQueue::AddED2KLink ( const CED2KFileLink* link, uint8_t category )

{
	CPartFile* file = NULL;

	if ( IsFileExisting ( link->GetHashKey() ) )
	{
		// Must be a shared file if we are to add hashes or sources

		if ( ( file = GetFileByID ( link->GetHashKey() ) ) == NULL )
		{
			return false;
		}
	}
	else
	{
		file = new CPartFile ( link );

		if ( file->GetStatus() == PS_ERROR )
		{
			delete file;
			return false;
		}

		AddDownload ( file, thePrefs::AddNewFilesPaused(), category );
	}

	if ( link->HasValidAICHHash() )
	{
		CAICHHashSet* hashset = file->GetAICHHashset();

		if ( !hashset->HasValidMasterHash() || ( hashset->GetMasterHash() != link->GetAICHHash() ) )
		{
			hashset->SetMasterHash ( link->GetAICHHash(), AICH_VERIFIED );
			hashset->FreeHashSet();
		}
	}

	const CED2KFileLink::CED2KLinkSourceList& list = link->m_sources;

	CED2KFileLink::CED2KLinkSourceList::const_iterator it = list.begin();

	for ( ; it != list.end(); ++it )
	{
		AddDebugLogLineM ( true, logDownloadQueue, CFormat ( wxT ( "Adding source for ED2K link: dest=%x" ) ) % it->addr.hashCode() );
		AddToResolve ( link->GetHashKey(), it->addr );//I2P:, it->port);
	}

	return true;
}


bool CDownloadQueue::AddED2KLink ( const CED2KServerLink* link )

{
	CServer *server = new CServer ( link->GetDest() );//I2P:Port(), Uint32_ttoStringIP( link->GetIP() ) );

	server->SetListName ( CFormat ( wxT ( "%x" ) ) % link->GetDest().hashCode() );//Uint32_ttoStringIP( link->GetIP() ) );

	theApp.serverlist->AddServer ( server );

	Notify_ServerAdd ( server );

	return true;
}


bool CDownloadQueue::AddED2KLink ( const CED2KServerListLink* link )

{
	theApp.serverlist->UpdateServerMetFromURL ( link->GetAddress() );

	return true;
}


void CDownloadQueue::ObserverAdded ( ObserverType* o )

{
	CObservableQueue<CPartFile*>::ObserverAdded ( o );

	EventType::ValueList list;

	{

		wxMutexLocker lock ( m_mutex );

		list.reserve ( m_filelist.size() );

		list.insert ( list.begin(), m_filelist.begin(), m_filelist.end() );
	}

	NotifyObservers ( EventType ( EventType::INITIAL, &list ), o );
}

bool	CDownloadQueue::KademliaSearchFile (
    uint32_t searchID,

    const Kademlia::CUInt128* pcontactID,
    //const Kademlia::CUInt128* pbuddyID,
    uint8_t type,
    const CI2PAddress & tcpdest,
    const CI2PAddress & udpdest,
    const CI2PAddress & serverdest,
    uint32_t WXUNUSED ( clientid )
)
{
	AddDebugLogLineM ( false, logDownloadQueue,
			   CFormat( wxT( "CDownloadQueue::KademliaSearchFile (type %i [1:not firewalled, 3:firewalled])" ) ) % type );

	//Safty measure to make sure we are looking for these sources
	CPartFile* temp = GetFileByKadFileSearchID ( searchID );

	if ( !temp )
	{
		AddDebugLogLineM ( false, logDownloadQueue, wxT ( "This is not the file we're looking for..." ) );
		return false;
	}

	//Do we need more sources?
	if ( temp->IsStopped() || thePrefs::GetMaxSourcePerFile() <= temp->GetSourceCount() )
	{
		AddDebugLogLineM ( false, logDownloadQueue, wxT ( "No more sources needed for this file" ) );
		return false;
	}

	//CI2PAddress ED2KDest = tcpdest;

	if ( theApp.ipfilter->IsFiltered ( tcpdest ) )
	{
		AddDebugLogLineM ( false, logDownloadQueue, CFormat ( wxT ( "Filtered source received from Kademlia : udpdest=%s" ) )
			%  udpdest.humanReadable() );
		return false;
	}

	if ( tcpdest == theApp.GetTcpDest() )
	{
		AddDebugLogLineM ( false, logDownloadQueue, wxT ( "Trying to add myself as source, ignore" ) );
		return false;
	}

	CUpDownClient* ctemp = NULL;

	switch ( type )
	{
		case 1:
		{
			//NonFirewalled users

			if ( !tcpdest )
			{
				AddDebugLogLineM ( false, logDownloadQueue, CFormat ( wxT ( "Ignored source (udpdest=%s) received from Kademlia, no tcp dest received" ) ) % udpdest.humanReadable() );
				return false;
			}

			if ( !IsGoodDest ( tcpdest ) )
			{
				AddDebugLogLineM ( false, logDownloadQueue,
					CFormat ( wxT ( "Ignored source (tcpdest=%s, udpdest=%s) received from Kademlia, filtered" ) )
					% tcpdest.humanReadable() % udpdest.humanReadable() );
				return false;
			}

			ctemp = new CUpDownClient ( tcpdest, serverdest, temp, false, true );

			ctemp->SetSourceFrom ( SF_KADEMLIA );

			ctemp->SetUDPDest ( udpdest );

			uint8_t cID[16];

			pcontactID->toByteArray ( cID );

			ctemp->SetUserHash ( CMD4Hash ( cID ) );

			break;
		}

		case 2:
		{
			AddDebugLogLineM ( false, logDownloadQueue, CFormat ( wxT ( "Ignored source (tcpdest=%s, udpdest=%s) received from Kademlia, wrong type (type=2)" ) )
			    % tcpdest.humanReadable() % udpdest.humanReadable() );
			//Don't use this type... Some clients will process it wrong..
			break;
		}

		case 3:
		{
			//This will be a firewaled client connected to Kad only.
			//We set the clientID to 1 as a Kad user only has 1 buddy.
			//ctemp = new CUpDownClient(tcp,1,0,0,temp,false, true);
			ctemp = new CUpDownClient ( tcpdest, CI2PAddress::null, temp, false, true );
			//The only reason we set the real IP is for when we get a callback
			//from this firewalled source, the compare method will match them.
			ctemp->SetSourceFrom ( SF_KADEMLIA );
			ctemp->SetUDPDest ( udpdest );
			uint8_t cID[16];
			pcontactID->toByteArray ( cID );
			ctemp->SetUserHash ( CMD4Hash ( cID ) );
			//pbuddyID->toByteArray ( cID );
			//ctemp->SetBuddyID ( cID );
			//ctemp->SetBuddyDest ( serverdest );
			//ctemp->SetBuddyPort(serverport);
			break;
		}
	}

	if ( ctemp )
	{
		AddDebugLogLineM ( false, logDownloadQueue, CFormat ( wxT ( "Happily adding a source (%s) type %d" ) )
		                   % ctemp->GetTCPDest().humanReadable() % type );
		return CheckAndAddSource ( temp, ctemp );
	}
	return false;
}

CPartFile* CDownloadQueue::GetFileByKadFileSearchID ( uint32_t id ) const

{

	wxMutexLocker lock ( m_mutex );

	for ( uint16_t i = 0; i < m_filelist.size(); ++i )
	{
		if ( id == m_filelist[i]->GetKadFileSearchID() )
		{
			return m_filelist[ i ];
		}
	}

	return NULL;

}

bool CDownloadQueue::DoKademliaFileRequest()

{
	return ( ( ::GetTickCount() - lastkademliafilerequest ) > KADEMLIAASKTIME );
}
