//
// This file is part of the imule Project.
//
// Copyright (c) 2003-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2002 Merkur ( devs@emule-project.net / http://www.emule-project.net )
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

#include "SafeFile.h"				// Interface declarations.
#include "MD4Hash.h"				// Needed for CMD4Hash
#include "ArchSpecific.h"			// Needed for ENDIAN_SWAP_*
#include "common/StringFunctions.h"		// Needed for unicode2char, etc.
#include "common/Format.h"
#include "kademlia/utils/UInt128.h"	// Needed for CUInt128
#include "i2p/CI2PAddress.h"			// Needed for CI2PAddress

#include <algorithm>		// Needed for std::min


#define CHECK_BOM(size, x) ((size >= 3)  && (x[0] == (char)0xEF) && (x[1] == (char)0xBB) && (x[2] == (char)0xBF))

const uint8_t BOMHeader[3] = {(uint8_t)0xEF, (uint8_t)0xBB, (uint8_t)0xBF};


CSafeIOException::CSafeIOException(const wxString& type, const wxString& desc)
	: CMuleException(wxT("SafeIO::") + type, desc) {}


CEOFException::CEOFException(const wxString& desc)
	: CSafeIOException(wxT("EOF"), desc) {}


CIOFailureException::CIOFailureException(const wxString& desc)
	: CSafeIOException(wxT("IOFailure"), desc) {}

CIOFailureException::CIOFailureException(const wxString& type, const wxString& desc)
	: CSafeIOException(wxT("IOFailure::") + type, desc) {}
	

///////////////////////////////////////////////////////////////////////////////
// CFileDataIO


CFileDataIO::~CFileDataIO()
{
}


bool CFileDataIO::Eof() const
{
	return GetPosition() >= GetLength();
}


void CFileDataIO::Read(void *buffer, size_t count)
{
	MULE_VALIDATE_PARAMS(buffer, wxT("Attempting to write to NULL buffer."));

	// Check that we read everything we wanted.
	if (doRead(buffer, count) == count) {
		return;
	}

	// To reduce potential system calls, we only do EOF checks when reads fail.
	if (Eof()) {
		throw CEOFException(wxT("Attempt to read past end of file."));
	} else {
		throw CIOFailureException(wxT("Read error, failed to read from file."));
	}
}


void CFileDataIO::Write(const void* buffer, size_t count)
{
	MULE_VALIDATE_PARAMS(buffer, wxT("Attempting to read from NULL buffer."));

	if (doWrite(buffer, count) != count) {
		throw CIOFailureException(wxT("Write error, failed to write to file."));
	}
}

void CFileDataIO::Reset()
{
	if (Seek(0) != 0)
	{
		MULE_VALIDATE_STATE(0, wxT("Reset resulted in invalid offset."));
	}
}


wxULongLong CFileDataIO::Seek(wxLongLong offset, wxSeekMode from)
{
	wxLongLong newpos = 0;
	switch (from) {
		case wxFromStart:
			newpos = offset;
			break;
			
		case wxFromCurrent:
			newpos = wxLongLong(GetPosition().GetValue()) + offset;
			break;
			
		case wxFromEnd:
			newpos = wxLongLong(GetLength().GetValue()) + offset;
			break;
			
		default:
			MULE_VALIDATE_PARAMS(false, wxT("Invalid seek-mode specified."));
	}
	
	MULE_VALIDATE_PARAMS(newpos >= wxLongLong(0),
			     wxT("Position after seeking would be less than zero!"));

	wxULongLong result = doSeek(wxULongLong(newpos.GetValue()));
	MULE_VALIDATE_STATE(result != wxULongLong(wxInvalidOffset),
			    wxT("Seeking resulted in invalid offset."));
	MULE_VALIDATE_STATE(result == wxULongLong(newpos.GetValue()),
			    CFormat(wxT("Target position (%s) and actual position (%s) disagree after seeking offset %s (seekmode=%d).\nLength=%s")) 
					% newpos.ToString() 
					% result.ToString()
					% offset.ToString()
					% from
					% GetLength().ToString()
			);
	
	return result;
}


uint8_t CFileDataIO::ReadUInt8()
{
	uint8_t value = 0;
	Read(&value, 1);

	return value;
}


uint16_t CFileDataIO::ReadUInt16()
{
	uint16_t value = 0;
	Read(&value, 2);
	
	return ENDIAN_SWAP_16(value);
}

CI2PAddress CFileDataIO::ReadAddress()
{
	CI2PAddress value;
	char bytes[CI2PAddress::validLength()];
	Read(bytes, CI2PAddress::validLength());
	value.readBytes( bytes );
	return value;
}

void CFileDataIO::WriteAddress(const CI2PAddress& value)
{
	uint16_t writeLen  = CI2PAddress::validLength();
	char bytes[writeLen];
	value.writeBytes(bytes);
	Write(bytes, writeLen);
}


uint32_t CFileDataIO::ReadUInt32()
{
	uint32_t value = 0;
	Read(&value, 4);
	
	return ENDIAN_SWAP_32(value);
}


wxULongLong CFileDataIO::ReadUInt64()
{
	wxULongLong value = 0;
	Read(&value, 8);
	
	return value;
}


CUInt128 CFileDataIO::ReadUInt128()
{
	CUInt128 value;
	uint32_t* data = (uint32_t*)value.getDataPtr();
	for (int i = 0; i < 4; i++) {
		data[i] = ReadUInt32();
	}

	return value;
}


CMD4Hash CFileDataIO::ReadHash()
{
	CMD4Hash value;
	Read(value.GetHash(), 16);

	return value;
}


wxString CFileDataIO::ReadString(bool bOptUTF8, uint8_t SizeLen, bool SafeRead)
{
	size_t readLen;
	switch (SizeLen) {
		case 2:	readLen = ReadUInt16();	break;
		case 4:	readLen = ReadUInt32();	break;
			
		default:
			MULE_VALIDATE_PARAMS(false, wxT("Invalid SizeLen value in ReadString"));
	}	

	if (SafeRead) {
		readLen = (size_t) std::min<wxULongLong>(wxULongLong(readLen),
			   GetLength() - GetPosition()).GetValue();
	}
	
	return ReadOnlyString(bOptUTF8, readLen);
}


wxString CFileDataIO::ReadOnlyString(bool bOptUTF8, size_t raw_len)
{
	// We only need to set the the NULL terminator, since we know that
	// reads will either succeed or throw an exception, in which case
	// we wont be returning anything
	char val[raw_len + 1];
	val[raw_len] = 0;
		
	Read(val, raw_len);
	wxString str;
	
	if (CHECK_BOM(raw_len, val)) {
		// This is a UTF8 string with a BOM header, skip header.
		str = UTF82unicode(val + 3);
	} else {
		if (bOptUTF8) {
			str = UTF82unicode(val);
			if (str.IsEmpty()) {
				// Fallback to system locale
				str = char2unicode(val);
			}
		} else {
			str = char2unicode(val);
		}
	}

	return str;
}


void CFileDataIO::WriteUInt8(uint8_t value)
{
	Write(&value, 1);
}


void CFileDataIO::WriteUInt16(uint16_t value)
{
	ENDIAN_SWAP_I_16(value);
	
	Write(&value, 2);
}


void CFileDataIO::WriteUInt32(uint32_t value)
{
	ENDIAN_SWAP_I_32(value);
	
	Write(&value, 4);
}


void CFileDataIO::WriteUInt64(wxULongLong value)
{
	Write(&value, 8);
}


void CFileDataIO::WriteUInt128(const Kademlia::CUInt128& value)
{
	for (int i = 0; i < 4; i++) {
		WriteUInt32(value.get32BitChunk(i));
	}
}


void CFileDataIO::WriteHash(const CMD4Hash& value)
{
	Write(value.GetHash(), 16);
}


void CFileDataIO::WriteString(const wxString& str, EUtf8Str eEncode, uint8_t SizeLen)
{
	switch (eEncode) {
		case utf8strRaw:
		case utf8strOptBOM: {
			Unicode2CharBuf s(unicode2UTF8(str));
			if (s) {
				WriteStringCore(s, eEncode, SizeLen);
				break;
			}
		}
		default: {
			Unicode2CharBuf s1(unicode2char(str));
			WriteStringCore(s1, utf8strNone, SizeLen);			
		}
	}
}


void CFileDataIO::WriteStringCore(const char *s, EUtf8Str eEncode, uint8_t SizeLen)
{
	size_t sLength = s ? strlen(s) : 0;
	size_t real_length = 0;
	if (eEncode == utf8strOptBOM) {
		real_length = sLength + 3; // For BOM header.
	} else {
		real_length = sLength;
	}

	switch (SizeLen) {
		case 0:
			// Don't write size.
			break;
			
		case 2:
			// Can't be higher than a uint16_t
			wxASSERT(real_length <= (size_t)0xFFFF);
			WriteUInt16((uint16_t)real_length);
			break;
			
		case 4:
			WriteUInt32(real_length);
			break;
			
		default:
			MULE_VALIDATE_PARAMS(false, wxT("Invalid length for string-length field."));
	}		
		
	// The BOM header must be written even if the string is empty.
	if (eEncode == utf8strOptBOM) {
		Write(BOMHeader, 3);
	}

	// Only attempt to write non-NULL strings.
	if (sLength) {
		// No NULL terminator is written since we explicitly specify the length
		Write(s, sLength);
	}
}

