//
// This file is part of the imule Project.
//
// Copyright (c) 2003-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2002 Merkur ( devs@emule-project.net / http://www.emule-project.net )
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

#ifndef SEARCHLIST_H
#define SEARCHLIST_H

#include "Types.h"		// Needed for uint8_t, uint16_t and uint32_t
#include "KnownFile.h"		// Needed for CAbstractFile
#include "Tag.h"		// Needed for TagList

#include <wx/thread.h>

#include <map>
#include <vector>
#include <list>

#include "i2p/CI2PAddress.h"

class CMemFile;
class CMD4Hash;
class CServer;
class CSearchList;
class CPacket;
	
namespace Kademlia {
	class CUInt128;
}

enum SearchType {
    KadSearch,
    LocalSearch,
	GlobalSearch
};

class CGlobalSearchThread : public wxThread 
{	
	int m_progress;
public:
	CGlobalSearchThread( CPacket *packet );
	~CGlobalSearchThread();

	int Progress() { return m_progress; }

private:
	CPacket* 	m_packet;

	virtual void* Entry();
};


class CSearchFile : public CAbstractFile
{
	friend class CPartFile;
public:
	/*I2P:CSearchFile(const CMemFile& in_data, bool bOptUTF8, long nSearchID, uint32_t nServerIP=0, uint16_t nServerPort=0, const wxString& pszDirectory = wxEmptyString, bool nKademlia = false);*/
	CSearchFile(CMemFile& in_data, long nSearchID, const CI2PAddress & nServerPort=CI2PAddress::null, const wxString& pszDirectory = wxEmptyString, bool nKademlia = false);
	
	virtual ~CSearchFile();

#ifdef CLIENT_GUI
	friend class CSearchListRem;
	CSearchFile(class CEC_SearchFile_Tag *);
	
	wxULongLong  	GetFileSize() { return m_nFileSize; }
#else

	wxULongLong		GetIntTagValue(uint8_t tagname) const;
	wxString	GetStrTagValue(uint8_t tagname) const;
	void		AddSources(uint32_t count, uint32_t count_complete);
	
	wxULongLong	GetFileSize() const;
#endif
	long	GetSearchID() const				{ return m_nSearchID; }
	//uint32_t	GetClientID() const				{ return m_nClientID; }
	//void	SetClientID(uint32_t nClientID)	{ m_nClientID = nClientID; }
	//uint16_t	GetClientPort() const			{ return m_nClientPort; }
	//void	SetClientPort(uint16_t nPort)		{ m_nClientPort = nPort; }
	CI2PAddress	GetClientDest() const			{ return m_nClientDest; }
	void	SetClientDest(const CI2PAddress & nDest)	{ m_nClientDest = nDest; }
	
	// This is used for some Kad checks on the child search results also, but 
	// as we don't implement any child search results, it's not used there :)
	bool	IsKademlia() const { return m_nKademlia; }
	// This can be used to raise a warning on downloading a file with 
	// i.e. 50 sources and none complete.
	int IsComplete() const;
	int IsComplete(uint32_t uSources, uint32_t uCompleteSources) const;
	
private:
	long		m_nSearchID;
	
	CI2PAddress		m_nClientDest;
	wxString	m_Directory;
	bool m_nKademlia;
};


class CSearchList
{
friend class CSearchListCtrl;
typedef std::vector<CSearchFile*> SearchList;

public:
	CSearchList();
	~CSearchList();
	void	Clear();

	wxString	StartNewSearch(uint32_t* nSearchID, 
								SearchType search_type, 
								const wxString& searchString, 
								const wxString& typeText,
								const wxString& extension, 
								wxULongLong min, 
								wxULongLong max, 
								uint32_t availability);

	void	ProcessSearchanswer(const char* in_packet, uint32_t size, CUpDownClient* Sender, bool* pbMoreResultsAvailable, const wxString& pszDirectory);
	void	ProcessSearchanswer(const char* packet, uint32_t size, const CI2PAddress & nServerDest/*uint32_t nServerIP, uint16_t nServerPort*/);
	//I2P:void	ProcessUDPSearchanswer(const CMemFile& packet, bool bOptUTF8, uint32_t nServerIP, uint16_t nServerPort);
	void	ProcessUDPSearchanswer(CMemFile& packet, const CI2PAddress & nServerDest);

	void	RemoveResults(long nSearchID);

	const SearchList GetSearchResults(long nSearchID);
	
	void	AddFileToDownloadByHash(const CMD4Hash& hash, uint8_t cat = 0);
	void	LocalSearchEnd();

	void StopGlobalSearch();
	
	void ClearThreadData(CGlobalSearchThread* thread = NULL) 
	{ 
		// The thread going down might not be the actual one.
		if (!thread || (thread == m_searchthread)) {
			m_searchthread = NULL; 
		}
		
		m_SearchInProgress = false;
		
	};
	
	bool IsGlobalSearch() { return m_searchthread != NULL; };

	bool SearchInProgress() { return m_SearchInProgress; }
	
	int Progress()
	{
		if ( m_searchthread ) {
			return m_searchthread->Progress();
		} else {
			return m_SearchInProgress ? 0 : 0xffff;
		}
	}
	
	bool KademliaSearchKeyword(uint32_t searchID, const Kademlia::CUInt128* pfileID, const wxString& name, wxULongLong size, const wxString& type, const TagList& taglist);
	
private:

	CMemFile *CreateSearchData(const wxString &searchString, wxString typeText,
				const wxString &extension, wxULongLong min, wxULongLong max, uint32_t availability, bool kad);

	CPacket* m_searchpacket;

	bool AddToList(CSearchFile* toadd, bool bClientResponse = false);

	typedef std::map<long, SearchList> ResultMap;
	
	ResultMap	m_Results;
	
	wxString	m_resultType;
	
	long		m_CurrentSearch;

	CGlobalSearchThread* m_searchthread;

	bool m_SearchInProgress;
};

#endif // SEARCHLIST_H
