#include "kadNodesListView.h"
#include "kademlia/routing/Contact.h"

#include "common/Format.h"
#include "Logger.h"

#include <wx/colour.h>
#include <wx/thread.h>
#include <algorithm>

typedef enum {TYPE, UDP_READABLE, TCP_READABLE, LAST_SEEN, KAD_ID_STRING, KAD_ID, DISTANCE, UDP_ID } columns_enum ;

static columns_enum __sort_order = TYPE ;
static bool __reverse = false ;
//static wxMutex contacts_mutex ;

struct Compare
{
	bool operator() ( const Kademlia::CContact & lhs, const Kademlia::CContact & rhs ) const
	{
		/*AddDebugLogLineM ( false, logKadPrefs,
				   CFormat ( wxT ( "sort --> Compare enter %x %x" ) ) % (int) &lhs % (int) &rhs  );
		*/
		bool inf = true ;
		try {
		switch ( __sort_order )
		{
			case TYPE         :
				inf = __reverse ? ( rhs.getType() < lhs.getType() )
					: ( lhs.getType() < rhs.getType() );
				break;
			case UDP_READABLE :
				inf = __reverse ? ( rhs.getUDPDest().humanReadable() < lhs.getUDPDest().humanReadable() )
				: ( lhs.getUDPDest().humanReadable() < rhs.getUDPDest().humanReadable() );
				break;
			case TCP_READABLE :
				inf = __reverse ? ( rhs.getTCPDest().humanReadable() < lhs.getTCPDest().humanReadable() )
				: ( lhs.getTCPDest().humanReadable() < rhs.getTCPDest().humanReadable() );
				break;
			case LAST_SEEN    :
				inf = __reverse ? ( rhs.getLastTypeUpdate() < lhs.getLastTypeUpdate() )
				: ( lhs.getLastTypeUpdate() < rhs.getLastTypeUpdate() );
				break;
			case KAD_ID_STRING:
			case KAD_ID       :
				inf = __reverse ? ( rhs.getClientID() < lhs.getClientID() )
				: ( lhs.getClientID() < rhs.getClientID() );
				break;
			case DISTANCE     :
				inf = __reverse ? ( rhs.getDistance() < lhs.getDistance() )
				: ( lhs.getDistance() < rhs.getDistance() );
				break;
			case UDP_ID       :
				inf = __reverse ? ( rhs.getUDPDest().toString() < lhs.getUDPDest().toString() ) :
					( lhs.getUDPDest().toString() < rhs.getUDPDest().toString() );
				break;
		}
		
		/*AddDebugLogLineM ( false, logKadPrefs,
				   CFormat ( wxT ( "sort --> Compare exit" ) )  );
		*/
		} catch (...)
		{
			AddDebugLogLineM ( true, logKadPrefs,
					   CFormat ( wxT ( "sort --> Compare EXCEPTION" ) )  );
		}
		return inf;
	}
};


KadNodesListView::KadNodesListView ( wxWindow* parent, wxWindowID id, const wxPoint& pos, const wxSize& size, long style, const wxValidator& validator, const wxString& name )
		:
		wxListView ( parent, id, pos, size, wxLC_REPORT | wxLC_VIRTUAL, validator, name )
{
	wxListItem itemCol;
	itemCol.SetText ( _ ( "node type" ) );
	itemCol.SetAlign ( wxLIST_FORMAT_CENTRE );
	InsertColumn ( 0, itemCol );

	itemCol.SetText ( _ ( "  id  " ) );
	itemCol.SetAlign ( wxLIST_FORMAT_CENTRE );
	InsertColumn ( 1, itemCol );

	itemCol.SetText ( _ ( "stream id" ) );
	itemCol.SetAlign ( wxLIST_FORMAT_CENTRE );
	InsertColumn ( 2, itemCol );

	itemCol.SetText ( _ ( "            last time seen            " ) );
	itemCol.SetAlign ( wxLIST_FORMAT_CENTRE );
	InsertColumn ( 3, itemCol );

	itemCol.SetText ( _ ( " kad id " ) );
	itemCol.SetAlign ( wxLIST_FORMAT_LEFT );
	InsertColumn ( 4, itemCol );

	itemCol.SetText ( _ ( " binary id " ) );
	itemCol.SetAlign ( wxLIST_FORMAT_LEFT );
	InsertColumn ( 5, itemCol );

	itemCol.SetText ( _ ( " distance " ) );
	itemCol.SetAlign ( wxLIST_FORMAT_LEFT );
	InsertColumn ( 6, itemCol );

	itemCol.SetText ( _ ( "   key   " ) );
	itemCol.SetAlign ( wxLIST_FORMAT_LEFT );
	InsertColumn ( 7, itemCol );

	SetTextColour ( *wxBLUE );

	for ( int i=0; i<8; i++ )
		SetColumnWidth ( i, wxLIST_AUTOSIZE_USEHEADER );

	SetItemCount ( 0 );
	Show();
}




void KadNodesListView::OnNodeEvent ( const GUIEvent & evt )
{
	Kademlia::CContact node = * ( Kademlia::CContact * ) evt.ptr_value ;
	delete ( Kademlia::CContact * ) evt.ptr_value ;
	
	CUInt128 kad_id = node . getClientID() ;

	switch ( evt.ID )
	{
		case KAD_NODE_ADD :
		{
			AddDebugLogLineM ( false, logKadPrefs, CFormat ( wxT ( "KadNodesListView::OnNodeEvent 1" ) ) );

			// is this node being added or updated ?
			bool updated = _contacts_map.count ( kad_id );

			if ( updated )
			{
				int i = _contacts_map[ kad_id ] ;
				AddDebugLogLineM ( false, logKadPrefs,
				                   CFormat ( wxT ( "Updating contact n°%ld" ) ) % i );
				_contacts_list[ i ] = node ;
			}
			else
			{
				AddDebugLogLineM ( false, logKadPrefs,
				                   CFormat ( wxT ( "Adding unknown contact n°%ld" ) ) % _contacts_list.size() );
				_contacts_map[ kad_id ] = _contacts_list . size() ;
				_contacts_list . push_back ( node ) ;
			}

			break;
		}
		case KAD_NODE_DELETE :
		{
			wxASSERT ( _contacts_map.count ( kad_id ) );
			unsigned int i = _contacts_map[ kad_id ];
			_contacts_list . erase ( _contacts_list.begin() + i ) ;
			_contacts_map  . erase ( kad_id );
			while ( i<_contacts_list.size() )
			{
				_contacts_map[ _contacts_list[i].getClientID() ]-- ;
				i++ ;
			}
			break;
		}
		default :
			wxASSERT_MSG ( false, wxT ( "Bad message received in KadNodesListView::OnNodeEvent" ) );
	}
	{
		//wxMutexLocker lock( contacts_mutex ) ;
		
		std::vector<Kademlia::CContact> old_list = _sorted_contacts_list ;
		_sorted_contacts_list = _contacts_list ;
		/*for (unsigned int i=0; i<_sorted_contacts_list.size(); i++)
		{
			_sorted_contacts_list.at(i).GetDataMutex().Lock();
		}*/
		std::sort ( _sorted_contacts_list.begin(), _sorted_contacts_list.end(), Compare() );
		/*for (unsigned int i=0; i<_sorted_contacts_list.size(); i++)
		{
			_sorted_contacts_list.at(i).GetDataMutex().Unlock();
		}*/

		if (_sorted_contacts_list.size() != old_list.size())
		{
			SetItemCount ( _sorted_contacts_list.size() );
		}
		for (unsigned int i=0 ; i<_sorted_contacts_list.size(); i++)
		{
			bool refresh_it = false ;
			if (i>=old_list.size())
			{
				refresh_it = true ;
			}
			else
			{
				Kademlia::CContact current = _sorted_contacts_list.at(i);
				Kademlia::CContact old     = old_list.at(i);
				if ( ( current.getType() != old.getType() )
							|| ( current.getUDPDest() != old.getUDPDest() )
						    || ( current.getTCPDest() != old.getTCPDest() )
						    || ( current.getLastTypeUpdate() != old.getLastTypeUpdate() )
						    || ( current.getClientID() != old.getClientID() )
						    || ( current.getDistance() != old.getDistance() ) )
				{
					refresh_it = true;
				}
			}
			if (refresh_it)
			{
				RefreshItem( i );
			}
		}
	}
}

wxString KadNodesListView::OnGetItemText ( long line, long col ) const
{
	//wxMutexLocker lock( contacts_mutex ) ;
	unsigned int item = (unsigned int) line, column = (unsigned int) col ;
	if (item >= _sorted_contacts_list.size()) return wxEmptyString ;
	
	Kademlia::CContact node = _sorted_contacts_list.at ( item ) ;

	switch ( column )
	{
		case TYPE         :  return wxString() << node.getType();
		case UDP_READABLE :  return node.getUDPDest().humanReadable() ;
		case TCP_READABLE :  return node.getTCPDest().humanReadable() ;
		case LAST_SEEN    :
		{
			time_t lastSeen = node.getLastTypeUpdate() ;
			return ( lastSeen==0 ? wxString() << _ ( "never" ) : wxDateTime ( lastSeen ).Format ( wxT ( "%c" ) ) ) ;
		}
		case KAD_ID_STRING:  return node.getClientIDString() ;
		case KAD_ID       :  return node.getClientID().toBinaryString().Mid ( 0,12 ) ;
		case DISTANCE     :  return node.getDistance().toBinaryString().Mid ( 0,12 ) ;
		case UDP_ID       :  return node.getUDPDest().toString() ;
		default           :  return wxString();
	}
}
wxListItemAttr * KadNodesListView::OnGetItemAttr ( long line ) const
{
	//wxMutexLocker lock( contacts_mutex ) ;
	static wxListItemAttr attr ;
	unsigned int item = (unsigned int) line ;

	if (item >= _sorted_contacts_list.size()) return NULL ;
	Kademlia::CContact node = _sorted_contacts_list.at ( item ) ;

	// set the colour according to type
	wxColour colour ( 0,0,0 ) ;
	switch ( node.getType() )
	{
		case 0 : colour = wxColour ( 0,0,255 ) ; break;
		case 1 : colour = wxColour ( 85,0,170 ); break;
		case 2 : colour = wxColour ( 170,0, 85 ) ; break;
		case 3 : colour = wxColour ( 255,0,  0 ) ; break;
		default: break;
	}
	attr.SetTextColour ( colour );
	return &attr ;
}

void KadNodesListView::OnColumnLClick ( wxListEvent & event )
{
	AddDebugLogLineM ( false, logKadPrefs,
			   CFormat ( wxT ( "ColumnLClick enter °%d" ) ) % (int) event.GetColumn() );
	wxASSERT ( event.GetColumn() >= 0 && event.GetColumn() <= UDP_ID );
	if ( __sort_order == event.GetColumn() )
	{
		__reverse = ! __reverse ;
	}
	else
	{
		__sort_order = ( columns_enum ) event.GetColumn() ;
	}
	{
		//wxMutexLocker lock( contacts_mutex ) ;
		AddDebugLogLineM ( false, logKadPrefs,
				   CFormat ( wxT ( "ColumnLClick sort °%d" ) ) % (int) event.GetColumn() );
		/*for (unsigned int i=0; i<_sorted_contacts_list.size(); i++)
		{
			_sorted_contacts_list.at(i).GetDataMutex().Lock();
		}*/
		std::sort ( _sorted_contacts_list.begin(), _sorted_contacts_list.end(), Compare() );
		/*for (unsigned int i=0; i<_sorted_contacts_list.size(); i++)
		{
			_sorted_contacts_list.at(i).GetDataMutex().Unlock();
		}*/
		AddDebugLogLineM ( false, logKadPrefs,
			   CFormat ( wxT ( "ColumnLClick refresh items °%d" ) ) % (int) event.GetColumn() );
		RefreshItems ( 0, _sorted_contacts_list.size()-1 ) ;
		AddDebugLogLineM ( false, logKadPrefs,
			   CFormat ( wxT ( "ColumnLClick exit °%d" ) ) % (int) event.GetColumn() );
	}
}


BEGIN_EVENT_TABLE ( KadNodesListView, wxListView )
	EVT_LIST_COL_CLICK ( -1, KadNodesListView::OnColumnLClick )
END_EVENT_TABLE()
