//
// C++ Implementation: CI2PAddress
//
// Description:
//
//
// Author: MKVore <mkvore@mail.i2p>, (C) 2007
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include <iostream>

#include "common/Format.h"
#include "Logger.h"
#include "ArchSpecific.h"
#include "b64codec.h"
#include "CI2PAddress.h"

static const char INVALID_TAG 		= 0;
static const char VALID_TAG 		= 1<<0;
static const char ALIAS_TAG 		= 1<<1;
static const char COMPLETE_FILE_TAG 	= 1<<2;
static const char INCOMPLETE_FILE_TAG 	= 1<<3;

const CI2PAddress CI2PAddress::null((const char)0);
const CI2PAddress CI2PAddress::complete_file	(COMPLETE_FILE_TAG);
const CI2PAddress CI2PAddress::incomplete_file	(INCOMPLETE_FILE_TAG);

uint16_t CI2PAddress::validLength()
{
	return PUBKEY_LEN;
}

CI2PAddress::CI2PAddress (const char & typetag)
{
	init();
	m_type     = typetag;
}

void CI2PAddress::init()
{
	memset(m_b64, 0, BASE64_PUBKEY_LEN+1);
	memset(m_bytes, 0, PUBKEY_LEN);
	m_alias    = wxEmptyString;
	m_hashCode = 0;
	m_type     = 0;
}

CI2PAddress::CI2PAddress( const sam_pubkey_t dest )
{
	base64::decoder dec;
	init();
	if (!dest) {
		AddDebugLogLineM(false, logAddress, wxString(wxT("Initializing CI2PAddress from NULL destination")));
		return;
	}
	try {
		memcpy( m_b64, dest, BASE64_PUBKEY_LEN );
		dec.decode(m_b64, BASE64_PUBKEY_LEN, m_bytes);
		setHashCode();
		m_type = VALID_TAG;
		
		AddDebugLogLineM(false, logAddress, CFormat(wxT("Initializing valid address from char array. bytes: %d"))
				% BASE64_PUBKEY_LEN);
		return;
	} catch (...) {
		AddDebugLogLineM(true, logAddress, CFormat( wxT("Initializing CI2PAddress from char array failed ... \n") ) );
		memset(m_b64, 0, BASE64_PUBKEY_LEN);
		memset(m_bytes, 0, PUBKEY_LEN);
		m_type = 0;
		return;
	}
}

CI2PAddress::CI2PAddress(const CI2PAddress & src)
{
	(*this) = src;
}

CI2PAddress& CI2PAddress::operator=( const CI2PAddress & src )
{
    init();
	memcpy( m_b64, src.m_b64, BASE64_PUBKEY_LEN );
	memcpy( m_bytes, src.m_bytes, PUBKEY_LEN );
	m_hashCode = src.m_hashCode;
	m_type = src.m_type;
	m_alias = src.m_alias.c_str();
	return (*this);
}
	
CI2PAddress CI2PAddress::fromString(wxString s)
{
	CI2PAddress ans = null;
	AddDebugLogLineM(false, logAddress, wxString(wxT("from string ")) << wxT(" <-- ") << s );
	ans = CI2PAddress( (const char *) s.ToAscii() );
	return ans;
}

void CI2PAddress::setAlias(wxString alias) 
{
	m_alias=alias.c_str();
	if (!m_alias.IsEmpty())
		m_type |= ALIAS_TAG ;
}


bool CI2PAddress::isValid() const
{
	return ( m_type & VALID_TAG );
}

bool CI2PAddress::operator!() const
{
	return !(isValid());
}


const char * CI2PAddress::toBase64CharArray() const 
{
	return m_b64 ;
}

bool CI2PAddress::operator== (const CI2PAddress & other) const
{
	if (isValid())
		if (other.isValid())
			return memcmp(m_bytes, other.m_bytes, PUBKEY_LEN)==0;
		else
			return false;
	if (other.isValid())
		return false;
	if ( (m_type & ALIAS_TAG) & other.m_type )
		return m_alias==other.m_alias;
	return m_type==other.m_type;
}

bool CI2PAddress::operator<  (const CI2PAddress & other) const
{
	return (m_type<other.m_type 
			|| m_alias<other.m_alias 
			|| memcmp(m_bytes, other.m_bytes, PUBKEY_LEN)<0 );
}

uint32_t CI2PAddress::size() const 
{
	return PUBKEY_LEN;
}

uint32_t CI2PAddress::writeBytes ( char * buffer ) const
{
	memcpy( buffer, m_bytes, PUBKEY_LEN );
	return PUBKEY_LEN;
}


uint32_t CI2PAddress::readBytes ( char * buffer )
{
	base64::encoder enc;
	m_type = INVALID_TAG ;
	try {
		if (memcmp(buffer,null.m_bytes,PUBKEY_LEN)!=0)
			m_type = VALID_TAG;
		
		if (m_type & VALID_TAG)
		{
			memcpy( m_bytes, buffer, PUBKEY_LEN );
			enc.encode(m_bytes, PUBKEY_LEN, m_b64);
			setHashCode();
			AddDebugLogLineM(false, logAddress, CFormat(wxT("read : dest=%x")) % hashCode());
		}
		else
		{
			memset( m_bytes, 0, PUBKEY_LEN);
			memset( m_b64, 0, BASE64_PUBKEY_LEN);
		}
		return PUBKEY_LEN;
	}
	catch (...)
	{
		AddDebugLogLineM(false, logAddress, CFormat(wxT("read : %x (failed with Unhandled exception)"))
			%  hashCode());
		throw;
	}
	(*this) = CI2PAddress::null;
	return 0;
}


void CI2PAddress::setHashCode()
{
	m_hashCode = 0;
	for (uint16_t i=0; i<PUBKEY_LEN; i++)
		m_hashCode += (unsigned char) m_bytes[i];
}

wxString CI2PAddress::humanReadable() const
{
	return CFormat(wxT("%lx")) % hashCode();
}
