//
// This file is part of the imule Project.
//
// Copyright (c) 2003-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2002 Merkur ( devs@emule-project.net / http://www.emule-project.net )
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

#include "imule.h"		// Needed for theApp
#include "ClientList.h"		// Interface declarations.
#include "ClientTCPSocket.h"	// Needed for CClientTCPSocket
#include "DownloadQueue.h"	// Needed for CDownloadQueue
#include "UploadQueue.h"	// Needed for CUploadQueue
#include "IPFilter.h"		// Needed for CIPFIlter
#include "updownclient.h"	// Needed for CUpDownClient
#include "OPCodes.h"
#include "GetTickCount.h"	// Needed for GetTickCount()
#include "OtherFunctions.h"	// Needed for IP_FROM_GUI_ID and PORT_FROM_GUI_ID
#include "ServerConnect.h"	// Needed for theApp.serverconnect
#include "Preferences.h"	// Needed for thePrefs
#include "Statistics.h"		// Needed for theStats
#include "GuiEvents.h"          // Needed for Nofify_... macros
#include "Logger.h"
#include <common/Format.h>

#include <algorithm>

#include "kademlia/routing/Contact.h"

#include "kademlia/kademlia/Kademlia.h"
#include "kademlia/kademlia/Prefs.h"
#include "kademlia/kademlia/Search.h"
#include "kademlia/net/KademliaUDPListener.h"

#include "i2p/CI2PAddress.h"

/**
 * CDeletedClient Class
 *
 * This class / list is a bit overkill, but currently needed to avoid any
 * exploit possibility. It will keep track of certain clients attributes
 * for 2 hours, while the CUpDownClient object might be deleted already.
 * Currently saves: IP, Port, UserHash.
 */

class CDeletedClient
{
	public:
		CDeletedClient ( CUpDownClient* pClient )
		{
			m_dwInserted = ::GetTickCount();
			DestAndHash desthash = { pClient->GetTCPDest(), pClient->GetCreditsHash() };
			m_ItemsList.push_back ( desthash );
		}

		struct DestAndHash
		{
			CI2PAddress nDest;
			void* pHash;
		};

		typedef std::list<DestAndHash> DaHList;
		DaHList	m_ItemsList;
		uint32_t	m_dwInserted;
};



CClientList::CClientList()
		: m_deadSources ( true )
{
	m_dwLastBannCleanUp = 0;
	m_dwLastTrackedCleanUp = 0;
	m_dwLastClientCleanUp = 0;
	//m_pBuddy = NULL;
	//m_nBuddyStatus = Disconnected;
#ifdef __WXDEBUG__
	m_delete_queue_closed = false;
#endif
}


CClientList::~CClientList()
{
	std::map<CI2PAddress, CDeletedClient*>::iterator it = m_trackedClientsList.begin();

	for ( ; it != m_trackedClientsList.end(); ++it )
	{
		delete it->second;
	}

	m_trackedClientsList.clear();

	wxASSERT ( m_clientList.empty() );
	wxASSERT ( m_delete_queue.empty() );
}


void CClientList::AddClient ( CUpDownClient* toadd )
{
	// Ensure that only new clients can be added to the list

	if ( toadd->GetClientState() == CS_NEW )
	{
		// Update the client-state
		toadd->m_clientState = CS_LISTED;

		Notify_ClientCtrlAddClient ( toadd );

		// We always add the ID/ptr pair, regardles of the actual ID value
		m_clientList.insert ( IDMapPair ( toadd->GetTCPDest(), toadd ) );

		// We only add the hash if it is valid
		if ( toadd->HasValidHash() )
		{
			m_hashList.insert ( HashMapPair ( toadd->GetUserHash(), toadd ) );
		}

		toadd->UpdateStats();
	}
}


void CClientList::AddToDeleteQueue ( CUpDownClient* client )
{
	//RemoveFromKadList ( client );

	// We have to remove the client from the list immediatly, to avoid it getting
	// found by functions such as AttachToAlreadyKnown and GetClientsFromIP,
	// however, if the client isn't on the clientlist, then it is safe to delete
	// it right now. Otherwise, push it onto the queue.

	if ( RemoveFromList ( client ) )
	{
		// Also remove the hash entry
		RemoveHashFromList ( client );

		wxASSERT ( !m_delete_queue_closed );
		m_delete_queue.push_back ( client );
	}
	else
	{
		delete client;
	}
}


void CClientList::UpdateClientDest/*IP*/ ( CUpDownClient* client, const CI2PAddress & newDest/*I2P:uint32_t newIP*/ )

{
	// Sanity check

	if ( ( client->GetClientState() != CS_LISTED ) || ( client->GetTCPDest() == newDest/*I2P:client->GetIP() == newIP*/ ) )
		return;

	// Remove the old IP entry
	RemoveFromList ( client );

	// Add the new entry
	m_clientList.insert ( IDMapPair ( newDest/*I2P:newIP*/, client ) );
}


void CClientList::UpdateClientHash ( CUpDownClient* client, const CMD4Hash& newHash )

{
	// Sanity check

	if ( ( client->GetClientState() != CS_LISTED ) || ( client->GetUserHash() == newHash ) )
		return;


	// Remove the old entry
	RemoveHashFromList ( client );

	// And add the new one if valid
	if ( !newHash.IsEmpty() )
	{
		m_hashList.insert ( HashMapPair ( newHash, client ) );
	}
}


bool CClientList::RemoveFromList ( CUpDownClient* client )

{
	bool result = false;

	// First remove the ID entry
	std::pair<IDMap::iterator, IDMap::iterator> range = m_clientList.equal_range ( client->GetTCPDest() );

	for ( ; range.first != range.second; ++range.first )
	{
		if ( client == range.first->second )
		{
			/* erase() will invalidate the iterator, but we're not using it anymore
			    anyway (notice the break;) */
			m_clientList.erase ( range.first );
			result = true;

			break;
		}
	}

	return result;
}


void CClientList::RemoveHashFromList ( CUpDownClient* client )
{
	// Nothing to remove

	if ( !client->HasValidHash() )
	{
		return;
	}

	// Find all items with the specified hash
	std::pair<HashMap::iterator, HashMap::iterator> range = m_hashList.equal_range ( client->GetUserHash() );

	for ( ; range.first != range.second; ++range.first )
	{
		if ( client == range.first->second )
		{
			/* erase() will invalidate the iterator, but we're not using it anymore
			    anyway (notice the break;) */
			m_hashList.erase ( range.first );
			break;
		}
	}
}


CUpDownClient* CClientList::FindMatchingClient ( CUpDownClient* client )
{
	typedef std::pair<IDMap::const_iterator, IDMap::const_iterator> IDMapIteratorPair;
	wxCHECK ( client, NULL );

	const CI2PAddress userDest = client->GetTCPDest();

	const CI2PAddress userKadDest = client->GetUDPDest();


	if ( userDest.isValid() or userKadDest.isValid() )
	{
		// Check by dest

		struct { const IDMap& map; CI2PAddress value; }

		toCheck[] = {
		                { m_clientList, userDest }
		            };

		for ( size_t i = 0; i < itemsof ( toCheck ); ++i )
		{
			if ( !toCheck[i].value )
			{
				// We may not have both (or any) of these values.
				continue;
			}

			IDMapIteratorPair range = toCheck[i].map.equal_range ( toCheck[i].value );

			if ( userDest.isValid() )
			{
				IDMap::const_iterator it = range.first;

				for ( ; it != range.second; ++it )
				{
					if ( userDest == it->second->GetTCPDest() )
					{
						return it->second;
					}
				}
			}

			if ( userKadDest.isValid() )
			{
				IDMap::const_iterator it = range.first;

				for ( ; it != range.second; ++it )
				{
					if ( userKadDest == it->second->GetUDPDest() )
					{
						return it->second;
					}
				}
			}
		}
	}


	// If anything else fails, then we look at hashes
	if ( client->HasValidHash() )
	{
		// Find all items with the specified hash
		std::pair<HashMap::iterator, HashMap::iterator> range = m_hashList.equal_range ( client->GetUserHash() );

		// Just return the first item if any

		if ( range.first != range.second )
		{
			return range.first->second;
		}
	}

	// Nothing found, must be a new client
	return NULL;
}


uint32_t CClientList::GetClientCount() const
{
	return m_clientList.size();
}


void CClientList::DeleteAll()

{
	m_hashList.clear();

	while ( !m_clientList.empty() )
	{
		IDMap::iterator it = m_clientList.begin();

		// Will call the removal of the item on this same class
		it->second->Disconnected ( wxT ( "Removed while deleting all from ClientList." ) );
		it->second->Safe_Delete();
	}

	// Clean up the clients now queued for deletion
#ifdef __WXDEBUG__
	m_delete_queue_closed = true;

#endif
	ProcessDeleteQueue();
}


bool CClientList::AttachToAlreadyKnown ( CUpDownClient** client, CClientTCPSocket* sender )
{
	CUpDownClient* tocheck = ( *client );

	CUpDownClient* found_client = FindMatchingClient ( tocheck );

	if ( tocheck == found_client )
	{
		// We found the same client instance (client may have sent more than one OP_HELLO). do not delete that client!
		return true;
	}

	if ( found_client != NULL )
	{
		if ( sender )
		{
			if ( found_client->GetSocket() )
			{
				if ( found_client->IsConnected()
				        && ( found_client->GetTCPDest() != tocheck->GetTCPDest() ) )
				{
					// if found_client is connected and has the IS_IDENTIFIED, it's safe to say that the other one is a bad guy

					if ( found_client->IsIdentified() )
					{
						AddDebugLogLineM ( false, logClient,
							CFormat ( wxT ( "%x :Client: %s (%x) Banreason: Another matching client with another tcp address, is already identified and connected" ) )
								% ( long ) tocheck % tocheck->GetUserName() % tocheck->GetTCPDest().hashCode() );
						tocheck->Ban();
						return false;
					}

					AddDebugLogLineM ( false, logClient, CFormat ( wxT ( "%x :WARNING! Found matching client, to a currently connected client: %s (%x) and %s (%x)" ) ) % ( long ) tocheck % tocheck->GetUserName() % tocheck->GetTCPDest().hashCode() % found_client->GetUserName() % found_client->GetTCPDest().hashCode() );

					return false;
				}

				found_client->GetSocket()->SetClient ( NULL );

				found_client->GetSocket()->Safe_Delete();
			}

			found_client->SetSocket ( sender );

			tocheck->SetSocket ( NULL );
		}

		*client = 0;

		tocheck->Safe_Delete();
		*client = found_client;
		return true;
	}

	return false;
}


/* I2P
CUpDownClient* CClientList::FindClientByIP( uint32_t clientip, uint16_t port )
{
	// Find all items with the specified ip
	std::pair<IDMap::iterator, IDMap::iterator> range = m_ipList.equal_range( clientip );

	for ( ; range.first != range.second; ++range.first ) {
		CUpDownClient* cur_client = range.first->second;
		// Check if it's actually the client we want
		if ( cur_client->GetUserPort() == port ) {
			return cur_client;
		}
	}

	return NULL;
}
*/

CUpDownClient* CClientList::FindClientByDest ( const CI2PAddress & clientdest )
{
	// Find all items with the specified ip
	std::pair<IDMap::iterator, IDMap::iterator> range = m_clientList.equal_range ( clientdest );

	for ( ; range.first != range.second; ++range.first )
	{
		CUpDownClient* cur_client = range.first->second;
		return cur_client;
	}

	return NULL;
}


bool CClientList::ComparePriorUserhash ( const CI2PAddress & nDest, void* pNewHash )

{
	std::map<CI2PAddress, CDeletedClient*>::iterator it = m_trackedClientsList.find ( nDest );

	if ( it != m_trackedClientsList.end() )
	{
		CDeletedClient* pResult = it->second;

		CDeletedClient::DaHList::iterator it2 = pResult->m_ItemsList.begin();

		for ( ; it2 != pResult->m_ItemsList.end(); ++it2 )
		{
			if ( it2->nDest == nDest )
			{
				if ( it2->pHash != pNewHash )
				{
					return false;
				}
				else
				{
					break;
				}
			}
		}
	}

	return true;
}


void CClientList::AddTrackClient ( CUpDownClient* toadd )

{
	std::map<CI2PAddress, CDeletedClient*>::iterator it = m_trackedClientsList.find ( toadd->GetTCPDest() );

	if ( it != m_trackedClientsList.end() )
	{
		CDeletedClient* pResult = it->second;

		pResult->m_dwInserted = ::GetTickCount();

		CDeletedClient::DaHList::iterator it2 = pResult->m_ItemsList.begin();

		for ( ; it2 != pResult->m_ItemsList.end(); ++it2 )
		{
			if ( it2->nDest == toadd->GetTCPDest() )
			{
				// already tracked, update
				it2->pHash = toadd->GetCreditsHash();
				return;
			}
		}

		// New client for that IP, add an entry
		CDeletedClient::DestAndHash desthash = { toadd->GetTCPDest(), toadd->GetCreditsHash() };

		pResult->m_ItemsList.push_back ( desthash );
	}
	else
	{
		m_trackedClientsList[ toadd->GetTCPDest() ] = new CDeletedClient ( toadd );
	}
}


uint16_t CClientList::GetClientsFromDest ( const CI2PAddress & dwDest )
{
	std::map<CI2PAddress, CDeletedClient*>::iterator it = m_trackedClientsList.find ( dwDest );

	if ( it != m_trackedClientsList.end() )
	{
		wxASSERT( it->second->m_ItemsList.size() <= (uint16_t)-1 );
		return (uint16_t) it->second->m_ItemsList.size();
	}
	else
	{
		return 0;
	}
}


/**
 * Remove clients from all queues
 */
void CClientList::ProcessDeleteQueue()

{
	// Delete pending clients

	while ( !m_delete_queue.empty() )
	{
		CUpDownClient* toremove = m_delete_queue.front();
		m_delete_queue.pop_front();

		// Doing what RemoveClient used to do. Just to be sure...
		theApp.uploadqueue->RemoveFromUploadQueue ( toremove );
		theApp.uploadqueue->RemoveFromWaitingQueue ( toremove );
		theApp.downloadqueue->RemoveSource ( toremove );

		Notify_ClientCtrlRemoveClient ( toremove );

		delete toremove;
	}
}


void CClientList::Process()
{
	const uint32_t cur_tick = ::GetTickCount();

	ProcessDeleteQueue();

	if ( m_dwLastBannCleanUp + BAN_CLEANUP_TIME < cur_tick )
	{
		m_dwLastBannCleanUp = cur_tick;

		ClientMap::iterator it = m_bannedList.begin();

		while ( it != m_bannedList.end() )
		{
			if ( it->second + CLIENTBANTIME < cur_tick )
			{
				ClientMap::iterator tmp = it++;

				m_bannedList.erase ( tmp );
				theStats::RemoveBannedClient();
			}
			else
			{
				++it;
			}
		}
	}


	if ( m_dwLastTrackedCleanUp + TRACKED_CLEANUP_TIME < cur_tick )
	{
		m_dwLastTrackedCleanUp = cur_tick;

		std::map<CI2PAddress, CDeletedClient*>::iterator it = m_trackedClientsList.begin();

		while ( it != m_trackedClientsList.end() )
		{
			std::map<CI2PAddress, CDeletedClient*>::iterator cur_src = it++;

			if ( cur_src->second->m_dwInserted + KEEPTRACK_TIME < cur_tick )
			{
				delete cur_src->second;
				m_trackedClientsList.erase ( cur_src );
			}
		}
	}


	CleanUpClientList();
}


void CClientList::AddBannedClient ( const CI2PAddress & nDest )
{
	m_bannedList[nDest] = ::GetTickCount();
	theStats::AddBannedClient();
}


/*bool CClientList::IsBannedClient(uint32_t dwIP)
{
	ClientMap::iterator it = m_bannedList.find( dwIP );

	if ( it != m_bannedList.end() ) {
		if ( it->second + CLIENTBANTIME > ::GetTickCount() ) {
			return true;
		} else {
			RemoveBannedClient(dwIP);
		}
	}
	return false;
}*/
bool CClientList::IsBannedClient ( const CI2PAddress & nDest )
{
	ClientMap::iterator it = m_bannedList.find ( nDest );

	if ( it != m_bannedList.end() )
	{
		if ( it->second + CLIENTBANTIME > ::GetTickCount() )
		{
			return true;
		}
		else
		{
			RemoveBannedClient ( nDest );
		}
	}

	return false;
}


/*void CClientList::RemoveBannedClient(uint32_t dwIP)
{
	m_bannedList.erase(dwIP);
	theStats::RemoveBannedClient();
}*/
void CClientList::RemoveBannedClient ( const CI2PAddress & nDest )
{
	m_bannedList.erase ( nDest );
	theStats::RemoveBannedClient();
}


void CClientList::FilterQueues()

{
	// Filter client list

	for ( IDMap::iterator it = m_clientList.begin(); it != m_clientList.end(); )
	{
		IDMap::iterator tmp = it++; // Don't change this to a ++it!

		if ( theApp.ipfilter->IsFiltered ( tmp->second->GetTCPDest() ) )
		{
			tmp->second->Disconnected ( wxT ( "Filtered by IPFilter" ) );
			tmp->second->Safe_Delete();
		}
	}
}


CClientList::SourceList	CClientList::GetClientsByHash ( const CMD4Hash& hash )
{
	SourceList results;

	// Find all items with the specified hash
	std::pair<HashMap::iterator, HashMap::iterator> range = m_hashList.equal_range ( hash );

	for ( ; range.first != range.second; ++range.first )
	{
		results.push_back ( range.first->second );
	}

	return results;
}


CClientList::SourceList	CClientList::GetClientsByDest ( const CI2PAddress & dest )

{
	SourceList results;

	// Find all items with the specified hash
	std::pair<IDMap::iterator, IDMap::iterator> range = m_clientList.equal_range ( dest );

	for ( ; range.first != range.second; range.first++ )
	{
		results.push_back ( range.first->second );
	}

	return results;
}


const CClientList::IDMap& CClientList::GetClientList()
{
	return m_clientList;
}


void CClientList::AddDeadSource ( const CUpDownClient* client )

{
	m_deadSources.AddDeadSource ( client );
}


bool CClientList::IsDeadSource ( const CUpDownClient* client )

{
	return m_deadSources.IsDeadSource ( client );
}

bool CClientList::SendMessage ( const CI2PAddress & client_dest /*id*/, const wxString& message )

{
	CUpDownClient* client = FindClientByDest ( client_dest );
	AddDebugLogLineM ( false, logClient, CFormat ( wxT ( "%x: Trying to Send Message." ) ) % ( long ) client );

	if ( client )
	{
		AddDebugLogLineM ( false, logClient, wxT ( "Sending." ) );
	}
	else
	{
		AddDebugLogLineM ( true, logClient,
		                   CFormat ( wxT ( "No client with address %x found in CClientList::SendMessage(). Creating" ) )
		                   %  client_dest.hashCode() );
		client = new CUpDownClient ( client_dest, CI2PAddress::null, NULL, true, true );
		AddClient ( client );
	}

	return client->SendMessage ( message );
}

void CClientList::SetChatState ( const CI2PAddress & client_dest, EChatState state )

{
//void CClientList::SetChatState(uint64 client_dest, uint8_t state) {
	CUpDownClient* client = FindClientByDest ( client_dest );

	if ( client )
	{
		client->SetChatState ( state );
	}
}

/* Kad stuff */

void CClientList::RequestTCP ( Kademlia::CContact* contact )
{

	// uint32_t nContactIP = wxUINT32_SWAP_ALWAYS(contact->getIPAddress()); // I2P
	CI2PAddress nContactDest = contact->getTCPDest();
	// don't connect ourself
	/* I2P
	if (theApp.GetPublicIP() == nContactIP && thePrefs::GetPort() == contact->getTCPPort()) {
		return;
	}
	*/

	if ( theApp.GetTcpDest() == nContactDest )
	{
		return;
	}

	//CUpDownClient* pNewClient = FindClientByIP(nContactIP, contact->getTCPPort()); //I2P
	CUpDownClient* pNewClient = FindClientByDest ( nContactDest );

	if ( !pNewClient )
	{
		//TODO #warning Do we actually have to check friendstate here?
		// pNewClient = new CUpDownClient(contact->getTCPPort(), contact->getIPAddress(), 0, 0, NULL, false, true); //I2P
		pNewClient = new CUpDownClient ( contact->getTCPDest(), CI2PAddress::null, NULL, false, true );
	}

	//Add client to the lists to be processed.

	// pNewClient->SetKadPort(contact->getUDPPort()); //I2P
	pNewClient->SetUDPDest ( contact->getUDPDest() );

	//pNewClient->SetKadState ( KS_QUEUED_FWCHECK );

	//NOTE : not for I2P: AddToKadList ( pNewClient ); // This was a direct adding, but I like to check duplicates

	//This method checks if this is a dup already.
	AddClient ( pNewClient );
}

/* NOTE : not for I2P
void CClientList::RequestBuddy ( Kademlia::CContact* contact )

{

	CI2PAddress nContactDest = contact->getTCPDest();
	// Don't connect to ourself

	if ( theApp.GetPublicDest() == nContactDest )
	{
		return;
	}

	CUpDownClient* pNewClient = FindClientByDest ( nContactDest );

	if ( !pNewClient )
	{
		pNewClient = new CUpDownClient ( nContactDest, CI2PAddress::null, NULL, false, true );
	}

	//Add client to the lists to be processed.
	pNewClient->SetKadDest ( contact->getUDPDest() );

	pNewClient->SetKadState ( KS_QUEUED_BUDDY );

	byte ID[16];

	contact->getClientID().toByteArray ( ID );

	pNewClient->SetUserHash ( CMD4Hash ( ID ) );

	AddToKadList ( pNewClient );

	//This method checks if this is a dup already.
	AddClient ( pNewClient );

}

void CClientList::IncomingBuddy ( Kademlia::CContact* contact, Kademlia::CUInt128* buddyID )
{

	CI2PAddress nContactDest = contact->getTCPDest();
	//If imule already knows this client, abort this.. It could cause conflicts.
	//Although the odds of this happening is very small, it could still happen.

	if ( FindClientByDest ( nContactDest ) )
	{
		return;
	}

	// Don't connect ourself
	if ( theApp.GetPublicDest() == nContactDest )
	{
		return;
	}

	//Add client to the lists to be processed.
	CUpDownClient* pNewClient = new CUpDownClient ( nContactDest, CI2PAddress::null, NULL, false, true );

	pNewClient->SetKadDest ( contact->getUDPDest() );

	pNewClient->SetKadState ( KS_INCOMING_BUDDY );

	byte ID[16];

	contact->getClientID().toByteArray ( ID );

	pNewClient->SetUserHash ( CMD4Hash ( ID ) );

	buddyID->toByteArray ( ID );

	pNewClient->SetBuddyID ( ID );

	AddToKadList ( pNewClient );

	AddClient ( pNewClient );
}
*/

/*NOTE : not for I2P:
void CClientList::RemoveFromKadList ( CUpDownClient* torem )
{
	wxCHECK_RET ( torem, wxT ( "NULL pointer in RemoveFromKadList" ) );

	if ( m_KadSources.erase ( torem ) )
	{
		if ( torem == m_pBuddy )
		{
			m_pBuddy = NULL;
			Notify_ServerUpdateED2KInfo();
		}
	}

}

void CClientList::AddToKadList ( CUpDownClient* toadd )
{
	wxCHECK_RET ( toadd, wxT ( "NULL pointer in AddToKadList" ) );

	m_KadSources.insert ( toadd ); // This will take care of duplicates.
}
*/


void CClientList::CleanUpClientList()
{
	// We remove clients which are not needed any more by time
	// this check is also done on CUpDownClient::Disconnected, however it will not catch all
	// cases (if a client changes the state without beeing connected)
	//
	// Adding this check directly to every point where any state changes would be more effective,
	// is however not compatible with the current code, because there are points where a client has
	// no state for some code lines and the code is also not prepared that a client object gets
	// invalid while working with it (aka setting a new state)
	// so this way is just the easy and safe one to go (as long as imule is basically single threaded)

	const uint32_t cur_tick = ::GetTickCount();

	if ( m_dwLastClientCleanUp + CLIENTLIST_CLEANUP_TIME < cur_tick )
	{
		m_dwLastClientCleanUp = cur_tick;
		uint32_t cDeleted = 0;
		IDMap::iterator current_it = m_clientList.begin();

		while ( current_it != m_clientList.end() )
		{
			CUpDownClient* pCurClient = current_it->second;
			++current_it; // Won't be used till while loop again

			if (
			    (
			        pCurClient->GetUploadState() == US_NONE
			        || pCurClient->GetUploadState() == US_BANNED
			        && !pCurClient->IsBanned()
			    )
			    && pCurClient->GetRequestFile() == NULL       // because seed sources are added with DS_NONE
			    && pCurClient->GetDownloadState() == DS_NONE
			    && pCurClient->GetChatState() == MS_NONE
			    /*&& (
			        ( pCurClient->GetKadState() == KS_NONE )
			        || (
			            ( pCurClient->GetKadState() == KS_INCOMING_BUDDY )
			            &&
			            // We didn't receive the promised buddy on 10 min
			            ( ( ::GetTickCount() - pCurClient->GetCreationTime() ) > KADEMLIABUDDYTIMEOUT )
			        )
			    )*/
			    && pCurClient->GetSocket() == NULL
			)
			{
				cDeleted++;
				pCurClient->Disconnected ( wxT ( "Removed during ClientList cleanup." ) );
				pCurClient->Safe_Delete();
			}
			else
			{
				if ( ! ( pCurClient->GetUploadState() == US_NONE || pCurClient->GetUploadState() == US_BANNED && !pCurClient->IsBanned() ) )
				{
					AddDebugLogLineM ( false, logClient,
					                   CFormat ( wxT ( "Debug: Not deleted client %x with up state: %i " ) )
					                   % ( long int ) pCurClient % pCurClient->GetUploadState() );
				}

				if ( ! ( pCurClient->GetDownloadState() == DS_NONE ) )
				{
					AddDebugLogLineM ( false, logClient,
					                   CFormat ( wxT ( "Debug: Not deleted client %x with down state: %i " ) )
					                   % ( long int ) pCurClient % pCurClient->GetDownloadState() );
				}

				if ( ! ( pCurClient->GetChatState() == MS_NONE ) )
				{
					AddDebugLogLineM ( false, logClient,
					                   CFormat ( wxT ( "Debug: Not deleted client %x with chat state: %i " ) )
					                   % ( long int ) pCurClient % pCurClient->GetChatState() );
				}

				/*NOTE : I2P : no firewall
				if ( ! ( pCurClient->GetKadState() == KS_NONE ) )
				{
					AddDebugLogLineM ( false, logClient,
					                   CFormat ( wxT ( "Debug: Not deleted client %x with kad state: %i dest: %x time left: %i" ) )
					                   % ( long int ) pCurClient % pCurClient->GetKadState() % pCurClient->GetTCPDest().hashCode() % ( pCurClient->GetCreationTime() + KADEMLIABUDDYTIMEOUT -::GetTickCount() ) );
				}*/

				if ( ! ( pCurClient->GetSocket() == NULL ) )
				{
					AddDebugLogLineM ( false, logClient,
					                   CFormat ( wxT ( "Debug: Not deleted client %x: has socket" ) ) % ( long int ) pCurClient );
				}

				AddDebugLogLineM ( false, logClient,

				                   CFormat ( wxT ( "Debug: Not deleted client %x with kad version: %i" ) )
				                   % ( long int ) pCurClient % pCurClient->GetKadVersion() );
			}
		}

		AddDebugLogLineM ( false, logClient, wxString::Format ( wxT ( "Cleaned ClientList, removed %i not used known clients" ), cDeleted ) );
	}
}

